import pickle
from unittest.mock import PropertyMock, patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from numpy import dtype

import bluepysnap.edges as test_module
from bluepysnap.bbp import Synapse
from bluepysnap.circuit import Circuit
from bluepysnap.circuit_ids import CircuitEdgeId, CircuitEdgeIds, CircuitNodeId, CircuitNodeIds
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.utils import IDS_DTYPE

from utils import PICKLED_SIZE_ADJUSTMENT, TEST_DATA_DIR


class TestEdges:
    def setup_method(self):
        circuit = Circuit(str(TEST_DATA_DIR / "circuit_config.json"))
        self.test_obj = test_module.Edges(circuit)

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.EdgePopulation)
        with pytest.raises(BluepySnapError):
            self.test_obj["unknown"]

    def test_iter(self):
        assert sorted(self.test_obj) == ["default", "default2"]

    def test_population_names(self):
        assert self.test_obj.population_names == ["default", "default2"]

    def test_keys_names(self):
        assert list(self.test_obj.keys()) == ["default", "default2"]

    def test_values_population(self):
        values = list(self.test_obj.values())
        assert isinstance(values[0], test_module.EdgePopulation)
        assert values[0].name == "default"

        assert isinstance(values[1], test_module.EdgePopulation)
        assert values[1].name == "default2"

    def test_items(self):
        keys, values = zip(*self.test_obj.items())
        assert keys == ("default", "default2")
        assert isinstance(values[0], test_module.EdgePopulation)
        assert values[0].name == "default"

        assert isinstance(values[1], test_module.EdgePopulation)
        assert values[1].name == "default2"

    def test_size(self):
        assert self.test_obj.size == 8

    def test_property_names(self):
        assert self.test_obj.property_names == {
            "@dynamics:param1",
            "@source_node",
            "@target_node",
            "afferent_center_x",
            "afferent_center_y",
            "afferent_center_z",
            "afferent_section_id",
            "afferent_section_pos",
            "afferent_surface_x",
            "afferent_surface_y",
            "afferent_surface_z",
            "conductance",
            "delay",
            "efferent_center_x",
            "efferent_center_y",
            "efferent_center_z",
            "efferent_section_id",
            "efferent_section_pos",
            "efferent_surface_x",
            "efferent_surface_y",
            "efferent_surface_z",
            "other1",
            "other2",
            "syn_weight",
        }

    def test_ids(self):
        np.random.seed(0)
        # single edge ID --> CircuitEdgeIds return populations with the 0 id
        expected = CircuitEdgeIds.from_tuples([("default", 0), ("default2", 0)])
        tested = self.test_obj.ids(0)
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)
        tested = self.test_obj.ids(np.int64(0))
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)
        tested = self.test_obj.ids(np.uint64(0))
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)
        tested = self.test_obj.ids(np.int32(0))
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)
        tested = self.test_obj.ids(np.uint32(0))
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)

        # single edge ID list --> CircuitEdgeIds return populations with the 0 id
        expected = CircuitEdgeIds.from_tuples([("default", 0), ("default2", 0)])
        assert self.test_obj.ids([0]) == expected

        # default3 population does not exist and is asked explicitly
        with pytest.raises(BluepySnapError):
            ids = CircuitEdgeIds.from_arrays(["default", "default3"], [0, 3])
            self.test_obj.ids(ids)

        # seq of node ID --> CircuitEdgeIds return populations with the array of ids
        expected = CircuitEdgeIds.from_arrays(
            ["default", "default", "default2", "default2"], [0, 1, 0, 1]
        )
        tested = self.test_obj.ids([0, 1])
        assert tested == expected
        tested = self.test_obj.ids((0, 1))
        assert tested == expected
        tested = self.test_obj.ids(np.array([0, 1]))
        assert tested == expected

        ids = [CircuitEdgeId("default", 0), CircuitEdgeId("default", 1)]
        assert self.test_obj.ids(ids) == CircuitEdgeIds.from_dict({"default": [0, 1]})

        with pytest.raises(BluepySnapError):
            ids = [CircuitEdgeId("default", 0), ("default", 1)]
            self.test_obj.ids(ids)

        # Check operations on global ids
        ids = self.test_obj.ids([0, 1, 2, 3])
        assert ids.filter_population("default").append(ids.filter_population("default2")) == ids

        expected = CircuitEdgeIds.from_arrays(["default2", "default2"], [0, 1])
        assert ids.filter_population("default2").limit(2) == expected

        tested = self.test_obj.ids(sample=2)
        expected = CircuitEdgeIds.from_arrays(["default2", "default"], [2, 2], sort_index=False)
        assert tested == expected

        tested = self.test_obj.ids(limit=5)
        expected = CircuitEdgeIds.from_dict({"default": [0, 1, 2, 3], "default2": [0]})
        assert tested == expected

        with pytest.raises(BluepySnapError) as e:
            self.test_obj.ids({"afferent_center_i": (10, 11)})
        assert "Unknown edge properties: {'afferent_center_i'}" == e.value.args[0]

        tested = self.test_obj.ids({"afferent_center_x": (1110, 1110.5)})
        expected = CircuitEdgeIds.from_dict({"default": [0], "default2": [0]})
        assert tested == expected

        tested = self.test_obj.ids(
            {"afferent_center_x": (1111, 1112), "efferent_center_z": (2132, 2134)}
        )
        expected = CircuitEdgeIds.from_dict({"default": [2], "default2": [2]})
        assert tested == expected

        tested = self.test_obj.ids(
            {"$and": [{"@dynamics:param1": (0, 2)}, {"afferent_surface_x": (1211, 1211)}]}
        )
        expected = CircuitEdgeIds.from_dict({"default": [1], "default2": [1]})
        assert tested == expected

        tested = self.test_obj.ids({"$or": [{"@dynamics:param1": (0, 2)}, {"@source_node": [0]}]})
        expected = CircuitEdgeIds.from_dict({"default": [0, 1, 2], "default2": [0, 1, 2]})
        assert tested == expected

        tested = self.test_obj.ids({"population": ["default2"], "afferent_center_x": (1113, 1114)})
        expected = CircuitEdgeIds.from_dict({"default2": [3]})
        assert tested == expected

        tested = self.test_obj.ids({"population": ["default3"], "afferent_center_x": (1113, 1114)})
        expected = CircuitEdgeIds.from_arrays([], [])
        assert tested == expected

        tested = self.test_obj.ids({"population": ["default", "default2"], "@target_node": [1]})
        expected = CircuitEdgeIds.from_dict({"default": [1, 2, 3], "default2": [1, 2, 3]})
        assert tested == expected

    def test_get(self):
        with pytest.raises(BluepySnapError, match="You need to set edge_ids in get."):
            self.test_obj.get(properties=["other2"])

        ids = CircuitEdgeIds.from_dict({"default": [0, 1, 2, 3], "default2": [0, 1, 2, 3]})
        tested = self.test_obj.get(ids, None)
        assert tested == ids

        tested = self.test_obj.get(ids, properties=self.test_obj.property_names)
        tested = pd.concat(df for _, df in tested)

        assert len(tested) == 8
        assert len(list(tested)) == 24

        # put NaN for the undefined values : only values for default2 in dropna
        assert len(tested.dropna()) == 4

        # the index of the dataframe is indentical to the CircuitEdgeIds index
        pdt.assert_index_equal(tested.index, ids.index)
        tested2 = self.test_obj.get([0, 1, 2, 3], properties=self.test_obj.property_names)
        tested2 = pd.concat(df for _, df in tested2)
        pdt.assert_frame_equal(tested2, tested)

        # tested columns
        tested = self.test_obj.get(ids, properties=["other2", "other1", "@source_node"])
        expected = pd.DataFrame(
            {
                "other2": np.array([np.NaN, np.NaN, np.NaN, np.NaN, 10, 11, 12, 13], dtype=float),
                "other1": np.array(
                    [np.NaN, np.NaN, np.NaN, np.NaN, "A", "B", "C", "D"], dtype=object
                ),
                "@source_node": np.array([2, 0, 0, 2, 2, 0, 0, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default", 3),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "edge_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested[expected.columns], expected)

        tested = self.test_obj.get(
            CircuitEdgeIds.from_dict({"default2": [0, 1, 2, 3]}),
            properties=["other2", "other1", "@source_node"],
        )
        # correct ordering when setting the dataframe with the population dataframe
        expected = pd.DataFrame(
            {
                "other2": np.array([10, 11, 12, 13], dtype=np.int32),
                "other1": np.array(["A", "B", "C", "D"], dtype=object),
                "@source_node": np.array([2, 0, 0, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "edge_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        with pytest.raises(KeyError, match="'default'"):
            tested.loc[("default", 0)]

        tested = self.test_obj.get(
            CircuitEdgeIds.from_dict({"default": [0, 1, 2, 3]}),
            properties=["other2", "other1", "@source_node"],
        )
        expected = pd.DataFrame(
            {
                "@source_node": np.array([2, 0, 0, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default", 3),
                ],
                names=["population", "edge_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        tested = self.test_obj.get(ids, properties="@source_node")
        expected = pd.DataFrame(
            {
                "@source_node": np.array([2, 0, 0, 2, 2, 0, 0, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default", 3),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "edge_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        tested = self.test_obj.get(ids, properties="other2")
        expected = pd.DataFrame(
            {
                "other2": np.array([np.NaN, np.NaN, np.NaN, np.NaN, 10, 11, 12, 13], dtype=float),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default", 3),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "edge_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        with pytest.raises(BluepySnapError, match="Unknown properties required: {'unknown'}"):
            next(self.test_obj.get(ids, properties=["other2", "unknown"]))

        with pytest.raises(BluepySnapError, match="Unknown properties required: {'unknown'}"):
            next(self.test_obj.get(ids, properties="unknown"))

    def test_afferent_nodes(self):
        assert self.test_obj.afferent_nodes(0) == CircuitNodeIds.from_arrays(["default"], [2])
        assert self.test_obj.afferent_nodes(np.int64(0)) == CircuitNodeIds.from_arrays(
            ["default"], [2]
        )
        assert self.test_obj.afferent_nodes(np.uint64(0)) == CircuitNodeIds.from_arrays(
            ["default"], [2]
        )
        assert self.test_obj.afferent_nodes(np.int32(0)) == CircuitNodeIds.from_arrays(
            ["default"], [2]
        )
        assert self.test_obj.afferent_nodes(np.int32(0)) == CircuitNodeIds.from_arrays(
            ["default"], [2]
        )
        assert self.test_obj.afferent_nodes(
            CircuitNodeId("default", 0)
        ) == CircuitNodeIds.from_arrays(["default"], [2])
        assert self.test_obj.afferent_nodes([0, 1]) == CircuitNodeIds.from_dict({"default": [2, 0]})
        ids = CircuitNodeIds.from_dict({"default": [0, 1], "default2": [0, 1]})
        assert self.test_obj.afferent_nodes(ids) == CircuitNodeIds.from_dict({"default": [2, 0]})
        assert self.test_obj.afferent_nodes(0, unique=False) == CircuitNodeIds.from_arrays(
            ["default", "default"], [2, 2]
        )

        # use global mapping for nodes
        assert self.test_obj.afferent_nodes({"other1": "A"}) == CircuitNodeIds.from_arrays([], [])
        assert self.test_obj.afferent_nodes({"mtype": "L6_Y"}) == CircuitNodeIds.from_dict(
            {"default": [0, 2]}
        )

    def test_efferent_nodes(self):
        assert self.test_obj.efferent_nodes(0) == CircuitNodeIds.from_arrays(["default"], [1])
        assert self.test_obj.efferent_nodes(np.int64(0)) == CircuitNodeIds.from_arrays(
            ["default"], [1]
        )
        assert self.test_obj.efferent_nodes(np.uint64(0)) == CircuitNodeIds.from_arrays(
            ["default"], [1]
        )
        assert self.test_obj.efferent_nodes(np.int32(0)) == CircuitNodeIds.from_arrays(
            ["default"], [1]
        )
        assert self.test_obj.efferent_nodes(np.uint32(0)) == CircuitNodeIds.from_arrays(
            ["default"], [1]
        )
        assert self.test_obj.efferent_nodes(
            CircuitNodeId("default", 0)
        ) == CircuitNodeIds.from_arrays(["default"], [1])
        assert self.test_obj.efferent_nodes([0, 2]) == CircuitNodeIds.from_dict({"default": [0, 1]})
        ids = CircuitNodeIds.from_dict({"default": [0, 2]})
        assert self.test_obj.efferent_nodes(ids) == CircuitNodeIds.from_dict({"default": [1, 0]})
        assert self.test_obj.efferent_nodes(0, unique=False) == CircuitNodeIds.from_arrays(
            ["default", "default"], [1, 1]
        )

        # use global mapping for nodes
        assert self.test_obj.efferent_nodes({"other1": "A"}) == CircuitNodeIds.from_arrays([], [])
        assert self.test_obj.efferent_nodes({"mtype": "L6_Y"}) == CircuitNodeIds.from_dict(
            {"default": [0, 1]}
        )

    def test_pathway_edges(self):
        properties = [Synapse.AXONAL_DELAY]
        source = CircuitNodeIds.from_dict({"default": [0, 1]})
        target = CircuitNodeIds.from_dict({"default": [1, 2]})

        expected_index = CircuitEdgeIds.from_dict({"default": [1, 2], "default2": [1, 2]})
        tested = self.test_obj.pathway_edges(source=source, target=target, properties=properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [88.1862],
                    [52.1881],
                    [88.1862],
                    [52.1881],
                ],
                columns=properties,
                index=expected_index.index,
            ),
            check_dtype=False,
        )

        properties = [Synapse.SOURCE_NODE_ID, "other1"]
        expected_index = CircuitEdgeIds.from_dict({"default": [1, 2], "default2": [1, 2]})
        tested = self.test_obj.pathway_edges(source=source, target=target, properties=properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [0, np.nan],
                    [0, np.nan],
                    [0, "B"],
                    [0, "C"],
                ],
                columns=properties,
                index=expected_index.index,
            ),
            check_dtype=False,
        )

        # without the properties should return the CircuitEdgeIds
        assert self.test_obj.pathway_edges(source, target) == expected_index
        assert self.test_obj.pathway_edges(source, target, None) == expected_index

        # without the properties and the target
        assert self.test_obj.pathway_edges(source, None) == CircuitEdgeIds.from_dict(
            {"default": [1, 2], "default2": [1, 2]}
        )
        # without the properties and the source
        assert self.test_obj.pathway_edges(None, source) == CircuitEdgeIds.from_dict(
            {"default": [0, 1, 2, 3], "default2": [0, 1, 2, 3]}
        )

        # raise if both source and target are not set
        with pytest.raises(BluepySnapError):
            self.test_obj.pathway_edges(None, None, None)

        # test with simple CircuitNodeId
        properties = [Synapse.SOURCE_NODE_ID, Synapse.TARGET_NODE_ID]
        source = CircuitNodeId("default", 0)
        target = CircuitNodeId("default", 1)
        expected_index = CircuitEdgeIds.from_dict({"default": [1, 2], "default2": [1, 2]})
        tested = self.test_obj.pathway_edges(source=source, target=target, properties=properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [0, 1],
                    [0, 1],
                    [0, 1],
                    [0, 1],
                ],
                columns=properties,
                index=expected_index.index,
            ),
            check_dtype=False,
        )

        # use global mapping for nodes
        assert self.test_obj.pathway_edges(
            source={"mtype": "L6_Y"}, target={"mtype": "L2_X"}
        ) == CircuitEdgeIds.from_tuples([("default", 0), ("default2", 0)])

    def test_afferent_edges(self):
        # without the properties
        target = CircuitNodeIds.from_dict({"default": [0, 1]})
        assert self.test_obj.afferent_edges(target, None) == CircuitEdgeIds.from_dict(
            {"default": [0, 1, 2, 3], "default2": [0, 1, 2, 3]}
        )

        # with a single int
        expected = CircuitEdgeIds.from_dict({"default": [1, 2, 3], "default2": [1, 2, 3]})
        assert self.test_obj.afferent_edges(1, None) == expected

        # with a CircuitNodeId
        assert self.test_obj.afferent_edges(CircuitNodeId("default", 1), None) == expected

        properties = [Synapse.AXONAL_DELAY]
        tested = self.test_obj.afferent_edges(1, properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [88.1862],
                    [52.1881],
                    [11.1058],
                    [88.1862],
                    [52.1881],
                    [11.1058],
                ],
                columns=properties,
                index=expected.index,
            ),
            check_dtype=False,
        )

        # with an undefined other1 field for the population default
        properties = [Synapse.SOURCE_NODE_ID, "other1"]
        expected_index = CircuitEdgeIds.from_dict(
            {"default": [0, 1, 2, 3], "default2": [0, 1, 2, 3]}
        )
        tested = self.test_obj.afferent_edges(
            CircuitNodeIds.from_dict({"default": [0, 1]}), properties=properties
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [2, np.nan],
                    [0, np.nan],
                    [0, np.nan],
                    [2, np.nan],
                    [2, "A"],
                    [0, "B"],
                    [0, "C"],
                    [2, "D"],
                ],
                columns=properties,
                index=expected_index.index,
            ),
            check_dtype=False,
        )

    def test_efferent_edges(self):
        target = CircuitNodeIds.from_dict({"default": [2]})
        expected = CircuitEdgeIds.from_dict({"default": [0, 3], "default2": [0, 3]})
        assert self.test_obj.efferent_edges(target, None) == expected
        assert self.test_obj.efferent_edges(2, None) == expected
        assert self.test_obj.efferent_edges(CircuitNodeId("default", 2), None) == expected

        properties = [Synapse.AXONAL_DELAY]
        tested = self.test_obj.efferent_edges(2, properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [99.8945],
                    [11.1058],
                    [99.8945],
                    [11.1058],
                ],
                columns=properties,
                index=expected.index,
            ),
            check_dtype=False,
        )

        # with an undefined other1 field for the population default
        properties = [Synapse.TARGET_NODE_ID, "other1"]
        expected_index = CircuitEdgeIds.from_dict({"default": [0, 3], "default2": [0, 3]})

        tested = self.test_obj.efferent_edges(2, properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [0, np.nan],
                    [1, np.nan],
                    [0, "A"],
                    [1, "D"],
                ],
                columns=properties,
                index=expected_index.index,
            ),
            check_dtype=False,
        )

    def test_pair_edges(self):
        # no connection between 0 and 2
        assert self.test_obj.pair_edges(0, 2, None) == CircuitEdgeIds.from_arrays([], [])
        actual = self.test_obj.pair_edges(0, 2, [Synapse.AXONAL_DELAY])
        assert next(actual, None) is None

        assert self.test_obj.pair_edges(2, 0, None) == CircuitEdgeIds.from_tuples(
            [("default", 0), ("default2", 0)]
        )

        properties = [Synapse.AXONAL_DELAY]
        tested = self.test_obj.pair_edges(2, 0, properties)
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(
            tested,
            pd.DataFrame(
                [
                    [99.8945],
                    [99.8945],
                ],
                columns=properties,
                index=CircuitEdgeIds.from_tuples([("default", 0), ("default2", 0)]).index,
            ),
            check_dtype=False,
        )

    def test_iter_connections(self):
        ids = CircuitNodeIds.from_dict({"default": [0, 1, 2], "default2": [0, 1, 2]})
        # ordered by target
        expected = [
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0)),
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0)),
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1)),
        ]
        for i, tested in enumerate(self.test_obj.iter_connections(source=ids, target=ids)):
            assert tested == expected[i]

        for i, tested in enumerate(self.test_obj.iter_connections(source=None, target=ids)):
            assert tested == expected[i]

        # same but ordered by source
        expected = [
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0)),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1)),
        ]
        for i, tested in enumerate(self.test_obj.iter_connections(source=ids, target=None)):
            assert tested == expected[i]

        expected = [
            (
                CircuitNodeId("default", 2),
                CircuitNodeId("default", 0),
                CircuitEdgeIds.from_dict({"default": [0]}),
            ),
            (
                CircuitNodeId("default", 0),
                CircuitNodeId("default", 1),
                CircuitEdgeIds.from_dict({"default": [1, 2]}),
            ),
            (
                CircuitNodeId("default", 2),
                CircuitNodeId("default", 1),
                CircuitEdgeIds.from_dict({"default": [3]}),
            ),
            (
                CircuitNodeId("default", 2),
                CircuitNodeId("default", 0),
                CircuitEdgeIds.from_dict({"default2": [0]}),
            ),
            (
                CircuitNodeId("default", 0),
                CircuitNodeId("default", 1),
                CircuitEdgeIds.from_dict({"default2": [1, 2]}),
            ),
            (
                CircuitNodeId("default", 2),
                CircuitNodeId("default", 1),
                CircuitEdgeIds.from_dict({"default2": [3]}),
            ),
        ]
        for i, tested in enumerate(
            self.test_obj.iter_connections(source=ids, target=ids, return_edge_ids=True)
        ):
            assert tested == expected[i]

        expected = [
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0), 1),
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1), 2),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1), 1),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 0), 1),
            (CircuitNodeId("default", 0), CircuitNodeId("default", 1), 2),
            (CircuitNodeId("default", 2), CircuitNodeId("default", 1), 1),
        ]
        for i, tested in enumerate(
            self.test_obj.iter_connections(source=ids, target=ids, return_edge_count=True)
        ):
            assert tested == expected[i]

        with pytest.raises(BluepySnapError):
            next(
                self.test_obj.iter_connections(
                    ids, ids, return_edge_ids=True, return_edge_count=True
                )
            )

    def test_pickle(self, tmp_path):
        pickle_path = tmp_path / "pickle.pkl"

        # trigger some cached properties, to makes sure they aren't being pickeld
        self.test_obj.size
        self.test_obj.property_names

        with open(pickle_path, "wb") as fd:
            pickle.dump(self.test_obj, fd)

        with open(pickle_path, "rb") as fd:
            test_obj = pickle.load(fd)

        assert pickle_path.stat().st_size < 100 + PICKLED_SIZE_ADJUSTMENT
        assert test_obj.size == 8



import libsonata
import numpy.testing as npt
import pytest

import bluepysnap.input as test_module
from bluepysnap.exceptions import BluepySnapError

from utils import TEST_DATA_DIR


class TestSynapseReplay:
    def setup_method(self):
        simulation = libsonata.SimulationConfig.from_file(TEST_DATA_DIR / "simulation_config.json")
        self.test_obj = test_module.SynapseReplay(simulation.input("spikes_1"))

    def test_all(self):
        snap_attrs = {a for a in dir(self.test_obj) if not a.startswith("_")}
        libsonata_attrs = {a for a in dir(self.test_obj._instance) if not a.startswith("_")}

        # check that wrapped instance's public methods are available in the object
        assert snap_attrs.symmetric_difference(libsonata_attrs) == {"reader"}
        assert isinstance(self.test_obj.reader, libsonata.SpikeReader)

        for a in libsonata_attrs:
            assert getattr(self.test_obj, a) == getattr(self.test_obj._instance, a)

        npt.assert_almost_equal(self.test_obj.reader["default"].get(), [[0, 10.775]])

    def test_no_such_attribute(self):
        """Check that the attribute error is raised from the wrapped libsonata object."""
        with pytest.raises(AttributeError, match="libsonata._libsonata.SynapseReplay"):
            self.test_obj.no_such_attribute


def test_get_simulation_inputs():
    simulation = libsonata.SimulationConfig.from_file(TEST_DATA_DIR / "simulation_config.json")
    inputs = test_module.get_simulation_inputs(simulation)

    assert isinstance(inputs, dict)
    assert inputs.keys() == {"spikes_1", "current_clamp_1"}

    assert isinstance(inputs["spikes_1"], test_module.SynapseReplay)
    assert isinstance(inputs["current_clamp_1"], libsonata._libsonata.Linear)

    with pytest.raises(BluepySnapError, match="Unexpected type for 'simulation': str"):
        test_module.get_simulation_inputs("fail_me")



from unittest.mock import Mock, patch

import click
from click.testing import CliRunner

from bluepysnap.cli import cli

from utils import TEST_DATA_DIR


@patch("bluepysnap.schemas.validate_nodes_schema", Mock(return_value=[]))
@patch("bluepysnap.schemas.validate_edges_schema", Mock(return_value=[]))
@patch("bluepysnap.schemas.validate_circuit_schema", Mock(return_value=[]))
def test_cli_correct():
    runner = CliRunner()
    result = runner.invoke(cli, ["validate", str(TEST_DATA_DIR / "circuit_config.json")])
    assert result.exit_code == 0
    assert click.style("No Error: Success.", fg="green") in result.stdout


def test_cli_no_config():
    runner = CliRunner()
    result = runner.invoke(cli, ["validate"])
    assert result.exit_code == 2
    assert "Missing argument 'CONFIG_FILE'" in result.stdout



from pathlib import Path
from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

import bluepysnap.circuit_ids as test_module
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.utils import IDS_DTYPE

from utils import setup_tempdir


def _multi_index():
    return pd.MultiIndex.from_arrays([["a", "a", "b", "a"], [0, 1, 0, 2]])


class TestCircuitNodeId:
    def setup_method(self):
        self.test_obj = test_module.CircuitNodeId("pop", 1)

    def test_init(self):
        assert isinstance(self.test_obj, test_module.CircuitNodeId)
        assert isinstance(self.test_obj, tuple)

    def test_accessors(self):
        assert self.test_obj.population == "pop"
        assert self.test_obj.id == 1


class TestCircuitEdgeId:
    def setup_method(self):
        self.test_obj = test_module.CircuitEdgeId("pop", 1)

    def test_init(self):
        assert isinstance(self.test_obj, test_module.CircuitEdgeId)
        assert isinstance(self.test_obj, tuple)

    def test_accessors(self):
        assert self.test_obj.population == "pop"
        assert self.test_obj.id == 1


class TestCircuitNodeIds:
    @property
    def ids_cls(self):
        return test_module.CircuitNodeIds

    @property
    def id_cls(self):
        return test_module.CircuitNodeId

    @property
    def id_name(self):
        return "node_ids"

    def _circuit_ids(self, populations, ids):
        index = pd.MultiIndex.from_arrays([populations, ids])
        index.names = ["population", self.id_name]
        return index

    def setup_method(self):
        self.test_obj_unsorted = self.ids_cls(_multi_index(), sort_index=False)
        self.test_obj_sorted = self.ids_cls(_multi_index())

    def test_init(self):
        values = pd.MultiIndex.from_arrays([["a", "a", "b", "a"], [0, 1, 0, 2]])
        tested = self.ids_cls(values, sort_index=False)
        assert tested.index.names == ["population", self.id_name]
        npt.assert_equal(tested.index.values, values.values)

        tested = self.ids_cls(values, sort_index=True)
        assert tested.index.names == ["population", self.id_name]
        npt.assert_equal(tested.index.values, np.sort(values.values))

        with pytest.raises(BluepySnapError):
            self.ids_cls(1)

        assert isinstance(self.test_obj_sorted, self.ids_cls)

    def test_from_arrays(self):
        tested = self.ids_cls.from_arrays(["a", "b"], [0, 1])
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [0, 1]))

        # keep ids ordering
        tested = self.ids_cls.from_arrays(["a", "b"], [1, 0], sort_index=False)
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [1, 0]))

        # keep population ordering
        tested = self.ids_cls.from_arrays(["b", "a"], [0, 1], sort_index=False)
        pdt.assert_index_equal(tested.index, self._circuit_ids(["b", "a"], [0, 1]))

        # keep duplicates
        tested = self.ids_cls.from_arrays(["a", "a"], [0, 0])
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "a"], [0, 0]))
        assert tested.index.size == 2

        with pytest.raises(BluepySnapError):
            self.ids_cls.from_arrays(["a", "a", "a"], [0, 0])

        with pytest.raises(BluepySnapError):
            self.ids_cls.from_arrays(["a", "a"], [0, 0, 0])

    def test_create_ids(self):
        tested = self.ids_cls.from_dict({"a": [0]})
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a"], [0]))

        tested = self.ids_cls.from_dict({"a": [0, 1]})
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "a"], [0, 1]))

        tested = self.ids_cls.from_dict({"a": [0], "b": [0]})
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [0, 0]))

        tested = self.ids_cls.from_dict({"a": [0], "b": [1]})
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [0, 1]))

        # keep duplicates
        tested = self.ids_cls.from_dict({"a": [0, 0]})
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "a"], [0, 0]))
        assert tested.index.size == 2

    def test_from_tuples(self):
        tested = self.ids_cls.from_tuples([self.id_cls("a", 0), self.id_cls("b", 1)])
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [0, 1]))
        tested = self.ids_cls.from_tuples([("a", 0), ("b", 1)])
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "b"], [0, 1]))

    def test_copy(self):
        tested = self.test_obj_sorted.copy()
        assert self.test_obj_sorted is not tested
        assert self.test_obj_sorted.index is not tested.index
        assert self.test_obj_sorted == tested

    def test_len(self):
        assert len(self.test_obj_sorted) == len(_multi_index())

    def test__locate(self):
        tested = self.test_obj_sorted._locate("a")
        npt.assert_equal(tested, [0, 1, 2])
        tested = self.test_obj_unsorted._locate("a")
        npt.assert_equal(tested, [0, 1, 3])

        tested = self.test_obj_sorted._locate("b")
        npt.assert_equal(tested, [3])
        tested = self.test_obj_unsorted._locate("b")
        npt.assert_equal(tested, [2])

    def test_filter_population(self):
        tested = self.test_obj_sorted.filter_population("a")
        pdt.assert_index_equal(tested.index, self._circuit_ids(["a", "a", "a"], [0, 1, 2]))

        tested = self.test_obj_sorted.filter_population("b")
        pdt.assert_index_equal(tested.index, self._circuit_ids(["b"], [0]))

        tested = self.test_obj_sorted.copy()
        tested.filter_population("b", inplace=True)
        pdt.assert_index_equal(tested.index, self._circuit_ids(["b"], [0]))

    def test_get_populations(self):
        tested = self.test_obj_sorted.get_populations()
        npt.assert_equal(tested, ["a", "a", "a", "b"])
        tested = self.test_obj_unsorted.get_populations()
        npt.assert_equal(tested, ["a", "a", "b", "a"])

        tested = self.test_obj_sorted.get_populations(unique=True)
        npt.assert_equal(tested, ["a", "b"])
        tested = self.test_obj_unsorted.get_populations(unique=True)
        npt.assert_equal(tested, ["a", "b"])

    def test_get_ids(self):
        tested = self.test_obj_sorted.get_ids()
        npt.assert_equal(tested, [0, 1, 2, 0])
        npt.assert_equal(tested.dtype, IDS_DTYPE)
        tested = self.test_obj_unsorted.get_ids()
        npt.assert_equal(tested, [0, 1, 0, 2])
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        tested = self.test_obj_sorted.get_ids(unique=True)
        npt.assert_equal(tested, [0, 1, 2])
        tested = self.test_obj_unsorted.get_ids(unique=True)
        npt.assert_equal(tested, [0, 1, 2])

    def test_sort(self):
        obj = self.ids_cls(_multi_index(), sort_index=False)
        expected = pd.MultiIndex.from_arrays([["a", "a", "a", "b"], [0, 1, 2, 0]])
        npt.assert_equal(obj.sort().index.values, expected.values)
        obj.sort(inplace=True)
        npt.assert_equal(obj.index.values, expected.values)

    def test_append(self):
        other = self.ids_cls(pd.MultiIndex.from_arrays([["c", "b", "c"], [0, 5, 1]]))
        expected = self.ids_cls(
            self._circuit_ids(["a", "a", "b", "a", "c", "b", "c"], [0, 1, 0, 2, 0, 5, 1])
        )
        assert self.test_obj_sorted.append(other, inplace=False) == expected

        other = self.ids_cls(pd.MultiIndex.from_arrays([["a"], [0]]))
        expected = self.ids_cls(self._circuit_ids(["a", "a", "b", "a", "a"], [0, 1, 0, 2, 0]))
        assert self.test_obj_sorted.append(other, inplace=False) == expected

        test_obj = self.ids_cls(_multi_index())
        other = self.ids_cls(pd.MultiIndex.from_arrays([["c", "b", "c"], [0, 5, 1]]))
        test_obj.append(other, inplace=True)
        expected = self.ids_cls(
            self._circuit_ids(["a", "a", "b", "a", "c", "b", "c"], [0, 1, 0, 2, 0, 5, 1])
        )
        assert test_obj == expected

    def test_intersection(self):
        test_obj = self.ids_cls.from_tuples(_multi_index(), sort_index=False)
        other = self.ids_cls.from_tuples([("b", 0), ("a", 3), ("a", 2)], sort_index=False)
        expected = self.ids_cls.from_tuples([("a", 2), ("b", 0)])
        res = test_obj.intersection(other)

        # res should be sorted when inplace=False
        assert res == expected
        assert test_obj != expected

        res = test_obj.intersection(other, inplace=True)
        assert res is None

        # test_obj index should not be sorted when inplace=True
        assert test_obj != expected
        assert all(expected.index == test_obj.index.sort_values())

    def test_sample(self):
        with patch("numpy.random.choice", return_value=np.array([0, 3])):
            tested = self.test_obj_unsorted.sample(2, inplace=False)
            assert tested == self.ids_cls(self._circuit_ids(["a", "b"], [0, 0]))

        tested = self.test_obj_unsorted.sample(2, inplace=False)
        assert len(tested) == 2

        tested = self.test_obj_unsorted.sample(25, inplace=False)
        assert len(tested) == len(_multi_index())

        values = _multi_index()
        test_obj = self.ids_cls(values)
        assert len(test_obj) == 4
        test_obj.sample(1, inplace=True)
        assert len(test_obj) == 1
        assert len(self.ids_cls(pd.MultiIndex.from_arrays([[], []])).sample(2)) == 0

    def test_limit(self):
        tested = self.test_obj_sorted.limit(2, inplace=False)
        assert len(tested) == 2
        assert tested == self.ids_cls(self._circuit_ids(["a", "a"], [0, 1]))
        assert len(self.ids_cls(pd.MultiIndex.from_arrays([[], []])).limit(2)) == 0

    def test_unique(self):
        tested = self.ids_cls.from_dict({"a": [0, 0, 1], "b": [1, 2, 2]}).unique()
        expected = self.ids_cls.from_dict({"a": [0, 1], "b": [1, 2]})
        assert tested == expected
        assert len(self.ids_cls(pd.MultiIndex.from_arrays([[], []])).unique()) == 0

    def test_tolist(self):
        expected = [("a", 0), ("a", 1), ("b", 0), ("a", 2)]
        assert self.test_obj_unsorted.tolist() == expected

    def test_combined_operators(self):
        tested = self.test_obj_sorted.copy()
        tested.filter_population("a", inplace=True)
        tested.limit(3, inplace=True)
        assert tested == self.ids_cls(self._circuit_ids(["a", "a", "a"], [0, 1, 2]))

        tested = self.test_obj_sorted.copy()
        tested = tested.filter_population("a").limit(3)
        assert tested == self.ids_cls(self._circuit_ids(["a", "a", "a"], [0, 1, 2]))

    def test_equal(self):
        values = _multi_index()
        test_obj = self.ids_cls(values)
        other = self.ids_cls(values)
        assert test_obj == other

        diff_values = pd.MultiIndex.from_arrays([["a", "a", "b"], [0, 1, 0]])
        other = self.ids_cls(diff_values)
        assert test_obj != other

        # different object
        assert test_obj != 1
        # python2 complains
        assert not test_obj.__eq__(1)

        # same object
        test_obj = self.ids_cls(values)
        same_obj = test_obj
        assert test_obj == same_obj

    def test___call__(self):
        # the call is used so we can use the CircuitIds directly in a dataframe
        data = [0, 1, 2, 3]
        df = pd.DataFrame(data={"data": data}, index=_multi_index())
        # test_obj_unsorted = ['a', 'a', 'b', 'a'], [0, 1, 0, 2]
        assert df.loc[self.test_obj_unsorted, "data"].tolist() == data
        # test_obj_unsorted = ['a', 'a', 'a', 'b'], [0, 1, 2, 0] returned value should swap 2 and 3
        assert df.loc[self.test_obj_sorted, "data"].tolist() == [0, 1, 3, 2]

    def test_printing(self):
        tested = self.test_obj_unsorted.__repr__()
        class_name = self.test_obj_sorted.__class__.__name__
        expected = """{}([('a', 0),
            ('a', 1),
            ('b', 0),
            ('a', 2)],
           names=['population', '{}'])""".format(
            class_name, self.id_name
        )

        assert tested == expected
        assert repr(self.test_obj_sorted) == str(self.test_obj_sorted)

    def test_roundtrip(self):
        with setup_tempdir() as tmp_dir:
            output = Path(tmp_dir, "output.csv")
            self.test_obj_sorted.to_csv(str(output))
            new = self.ids_cls.from_csv(str(output))
            assert self.test_obj_sorted == new

    def test___iter__(self):
        assert list(self.test_obj_sorted) == [("a", 0), ("a", 1), ("a", 2), ("b", 0)]
        assert all(type(cid) == self.id_cls for cid in self.test_obj_sorted)

    def test___getitem__(self):
        assert self.test_obj_sorted[0] == self.id_cls("a", 0)


class TestCircuitEdgeIds(TestCircuitNodeIds):
    @property
    def ids_cls(self):
        return test_module.CircuitEdgeIds

    @property
    def id_cls(self):
        return test_module.CircuitEdgeId

    @property
    def id_name(self):
        return "edge_ids"



from pathlib import Path
from unittest.mock import Mock

import h5py
import libsonata
import numpy as np
import pytest

import bluepysnap.neuron_models as test_module
from bluepysnap.circuit import Circuit, CircuitConfig
from bluepysnap.circuit_ids import CircuitNodeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Node

from utils import TEST_DATA_DIR, copy_config, copy_test_data, create_node_population, edit_config


def test_invalid_model_type():
    """test that model type, other than 'biophysical' throws an error"""
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        config = CircuitConfig.from_config(config_copy_path).to_dict()
        nodes_file = config["networks"]["nodes"][0]["nodes_file"]
        with h5py.File(nodes_file, "r+") as h5f:
            grp_name = "nodes/default/0/model_type"
            data = h5f[grp_name][:]
            del h5f[grp_name]
            h5f.create_dataset(
                grp_name, data=["virtual"] * data.shape[0], dtype=h5py.string_dtype()
            )
        nodes = create_node_population(nodes_file, "default")
        with pytest.raises(BluepySnapError) as e:
            test_module.NeuronModelsHelper(Mock(), nodes)
        assert "biophysical node population" in e.value.args[0]


def test_get_invalid_node_id():
    nodes = create_node_population(str(TEST_DATA_DIR / "nodes.h5"), "default")
    config = CircuitConfig.from_config(TEST_DATA_DIR / "circuit_config.json").to_dict()
    test_obj = test_module.NeuronModelsHelper(nodes._properties, nodes)

    with pytest.raises(BluepySnapError) as e:
        test_obj.get_filepath("1")
    assert "node_id must be a int or a CircuitNodeId" in e.value.args[0]


def test_get_filepath_biophysical():
    nodes = create_node_population(str(TEST_DATA_DIR / "nodes.h5"), "default")
    config = CircuitConfig.from_config(TEST_DATA_DIR / "circuit_config.json").to_dict()
    test_obj = test_module.NeuronModelsHelper(nodes._properties, nodes)

    node_id = 0
    assert nodes.get(node_id, properties=Node.MODEL_TEMPLATE) == "hoc:small_bio-A"
    actual = test_obj.get_filepath(node_id)
    expected = Path(nodes._properties.biophysical_neuron_models_dir, "small_bio-A.hoc")
    assert actual == expected

    actual = test_obj.get_filepath(np.int64(node_id))
    assert actual == expected
    actual = test_obj.get_filepath(np.uint64(node_id))
    assert actual == expected
    actual = test_obj.get_filepath(np.int32(node_id))
    assert actual == expected
    actual = test_obj.get_filepath(np.uint32(node_id))
    assert actual == expected

    node_id = CircuitNodeId("default", 0)
    assert nodes.get(node_id, properties=Node.MODEL_TEMPLATE) == "hoc:small_bio-A"
    actual = test_obj.get_filepath(node_id)
    assert actual == expected

    node_id = CircuitNodeId("default", 2)
    assert nodes.get(node_id, properties=Node.MODEL_TEMPLATE) == "hoc:small_bio-C"
    actual = test_obj.get_filepath(node_id)
    expected = Path(nodes._properties.biophysical_neuron_models_dir, "small_bio-C.hoc")
    assert actual == expected


def test_absolute_biophysical_dir():
    with copy_test_data() as (circuit_dir, circuit_config):
        neuron_dir = circuit_dir / "biophysical_neuron_models"
        with h5py.File(str(circuit_dir / "nodes.h5"), "r+") as h5:
            template = [t.decode().split(":") for t in h5["nodes/default/0/model_template"]]
            template = [t[0] + ":" + str(neuron_dir / t[1]) for t in template]
            h5["nodes/default/0/model_template"][...] = template

        nodes = Circuit(circuit_config).nodes["default"]
        test_obj = test_module.NeuronModelsHelper(nodes._properties, nodes)

        for i, t in enumerate(template):
            assert str(test_obj.get_filepath(i)) == ".".join(t.split(":")[::-1])


def test_no_biophysical_dir():
    with copy_test_data() as (data_dir, circuit_config):
        with edit_config(circuit_config) as config:
            del config["components"]["biophysical_neuron_models_dir"]

        with pytest.raises(
            libsonata.SonataError,
            match="Node population .* is defined as 'biophysical' but does not define 'biophysical_neuron_models_dir'",
        ):
            Circuit(circuit_config).nodes["default"]



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Simulation input access."""
import libsonata

from bluepysnap import BluepySnapError


class SynapseReplay:
    """Wrapper class for libsonata.SynapseReplay to provide the reader as a property."""

    def __init__(self, instance):
        """Wrap libsonata SynapseReplay object.

        Args:
            instance (libsonata.SynapseReplay): instance to wrap
        """
        self._instance = instance

    def __dir__(self):
        """Provide wrapped SynapseReplay instance's public attributes in dir."""
        public_attrs_instance = {attr for attr in dir(self._instance) if not attr.startswith("_")}
        return list(set(super().__dir__()) | public_attrs_instance)

    def __getattr__(self, name):
        """Retrieve attributes from the wrapped object."""
        return getattr(self._instance, name)

    @property
    def reader(self):
        """Return a spike reader object for the instance."""
        return libsonata.SpikeReader(self.spike_file)


def get_simulation_inputs(simulation):
    """Get simulation inputs as a dictionary.

    Args:
        simulation (libsonata.SimulationConfig): libsonata Simulation instance

    Returns:
        dict: inputs with input names as keys and corresponding objects as values
    """

    def _get_input(name):
        """Helper function to wrap certain objects."""
        item = simulation.input(name)

        if item.module.name == "synapse_replay":
            return SynapseReplay(item)
        return item

    if isinstance(simulation, libsonata.SimulationConfig):
        return {name: _get_input(name) for name in simulation.list_input_names}

    raise BluepySnapError(f"Unexpected type for 'simulation': {simulation.__class__.__name__}")



"""Standalone module that validates Sonata circuit. See ``validate`` function.

The idea here is to not depend on libsonata if possible, so we can use this in all situations
"""
import logging
from pathlib import Path

import click
import h5py
import numpy as np
import pandas as pd

from bluepysnap import schemas
from bluepysnap.config import Parser
from bluepysnap.morph import EXTENSIONS_MAPPING
from bluepysnap.sonata_constants import DEFAULT_EDGE_TYPE, DEFAULT_NODE_TYPE
from bluepysnap.utils import load_json

L = logging.getLogger("brainbuilder")
MAX_MISSING_FILES_DISPLAY = 10


class Error:
    """Error used for reporting of validation errors."""

    FATAL = "FATAL"
    WARNING = "WARNING"
    INFO = "INFO"

    def __init__(self, level, message=None):
        """Error.

        Args:
            level (str): error level
            message (str|None): message
        """
        self.level = level
        self.message = message

    def __str__(self):
        """Returns only message by default."""
        return str(self.message)

    __repr__ = __str__

    def __eq__(self, other):
        """Two errors are equal if inherit from Error and their level, message are equal."""
        if not isinstance(other, Error):
            return False
        return self.level == other.level and self.message == other.message

    def __hash__(self):
        """Hash. Errors with the same level and message give the same hash."""
        return hash(self.level) ^ hash(self.message)


def fatal(message):
    """Shortcut for a fatal error.

    Args:
        message (str): text message

    Returns:
        Error: Error with level FATAL
    """
    return Error(Error.FATAL, message)


def _check_partial_circuit_config(config):
    return config.get("metadata", {}).get("status") == "partial"


def _check_components_dir(name, components):
    """Checks existence of directory within Sonata config 'components'.

    Args:
        name (str): components directory name
        components (dict): ref to config's components

    Returns:
        list: List of errors, empty if no errors
    """
    dirpath = components.get(name)
    if not dirpath or not Path(dirpath).is_dir():
        return [fatal(f'Invalid components "{name}": {dirpath}')]
    return []


def _check_files(name, files, level):
    """Checks for existence of files within an h5 group.

    Args:
        name (str): h5 group name
        files (Iterable): files to check for existence
        level (str): level of generated errors for missing files

    Returns:
        list: List of errors, empty if no errors
    """
    files = set(files)
    missing = []
    for f in sorted(files):
        if not f.is_file():
            missing.append(f)

        if len(missing) >= MAX_MISSING_FILES_DISPLAY:
            break

    if missing:
        filenames = "".join(f"\t{m.name}\n" for m in missing)

        return [
            Error(level, f"missing at least {len(missing)} files in group {name}:\n{filenames}")
        ]

    return []


def _print_errors(errors):
    """Some fancy errors printing."""
    colors = {Error.WARNING: "yellow", Error.FATAL: "red", Error.INFO: "green"}

    if not errors:
        print(click.style("No Error: Success.", fg=colors[Error.INFO]))

    for error in errors:
        print(click.style(error.level + ": ", fg=colors[error.level]) + str(error))


def _check_duplicate_populations(networks, key):
    """Check that that for key = nodes|edges, no duplicate populations names exists."""
    seen = set()
    errors = []
    for network in networks.get(key, {}):
        for population in network.get("populations", {}):
            if population in seen:
                errors.append(
                    fatal(f'Already have population "{population}" in config for type "{key}"')
                )
            seen.add(population)

    return errors


def _find_nodes_population(node_population_name, nodes):
    """Finds node population item in config.

    Args:
        node_population_name (str): name of node population
        nodes (list): "nodes" part of the resolved bluepysnap config

    Returns:
        dict/None: A dict item of "nodes" of the resolved bluepysnap config.
        None if it finds nothing.
    """
    for nodes_dict in nodes:
        if node_population_name in nodes_dict.get("populations", {}):
            return nodes_dict

    return None


def _get_group_name(group, parents=0):
    """Gets group name of h5 group.

    Args:
        group (h5py.Group): nodes group in nodes .h5 file
        parents (int): number of extra parents needed
    """
    return Path(*Path(group.name).parts[-(parents + 1) :])


def _get_model_template_file(model_template):
    """Resolves 'model_template' field of nodes group to a proper filename."""
    parts = model_template.split(":", 1)
    return parts[1] + "." + parts[0]


def _nodes_group_to_dataframe(group, population):
    """Transforms hdf5 population group to pandas DataFrame.

    Args:
        group: HDF5 nodes group
        population: HDF5 nodes population

    Returns:
        pd.DataFrame: dataframe with all group attributes
    """
    # TODO: remove multi-indexing (BBP only supports group '0')
    df = pd.DataFrame(population["node_type_id"][:], columns=["type_id"])
    size = df.size
    df["id"] = population["node_id"] if "node_id" in population else np.arange(size)
    df["group_id"] = population["node_group_id"] if "node_group_id" in population else 0
    df["group_index"] = (
        population["node_group_index"] if "node_group_index" in population else np.arange(size)
    )
    df = df[df["group_id"] == int(str(_get_group_name(group)))]

    for k, v in group.items():
        if k == "@library":
            continue
        if isinstance(v, h5py.Dataset):
            if v.dtype == h5py.string_dtype():
                df[k] = v.asstr()[:]
            else:
                df[k] = v[:]

    if "@library" in group:
        for k, v in group["@library"].items():
            if isinstance(v, h5py.Dataset):
                df[k] = pd.Categorical.from_codes(df[k], categories=v.asstr()[:])

    return df


def _check_bio_nodes_group(group_df, group, population, population_name):
    """Checks biophysical nodes group for errors.

    Args:
        group_df (pd.DataFrame): nodes group as a dataframe
        group (h5py.Group): nodes group in nodes .h5 file
        population (dict): a merged dictionary (current population and 'components' in config)
        population_name (str): name of the population

    Returns:
        list: List of errors, empty if no errors
    """
    L.debug("Check biophysical nodes group")

    errors = []

    group_name = _get_group_name(group, parents=1)

    morph_dirs = set()
    if "morphologies_dir" in population:
        dir_errors = _check_components_dir("morphologies_dir", population)
        errors += dir_errors
        if len(dir_errors) == 0:
            morph_dirs = {(population["morphologies_dir"], "swc")}

    if "alternate_morphologies" in population:
        for morph_type, morph_path in population["alternate_morphologies"].items():
            dir_errors = _check_components_dir(morph_type, population["alternate_morphologies"])
            errors += dir_errors
            if len(dir_errors) == 0:
                for extension, _type in EXTENSIONS_MAPPING.items():
                    if _type == morph_type:
                        morph_dirs |= {(morph_path, extension)}

    if "morphologies_dir" not in population and "alternate_morphologies" not in population:
        errors.append(
            fatal(
                "at least one of 'morphologies_dir' or 'alternate_morphologies' "
                f"must to be defined for 'biophysical' population '{population_name}'"
            )
        )

    if "morphology" in group_df.columns:
        for morph_path, extension in morph_dirs:
            L.debug("Checking morph files (%s): %s", extension, morph_path)

            errors += _check_files(
                f"morphology: {group_name}[{group.file.filename}]",
                (Path(morph_path, m + "." + extension) for m in group_df["morphology"].unique()),
                Error.WARNING,
            )

    if "biophysical_neuron_models_dir" in population:
        errors += _check_components_dir("biophysical_neuron_models_dir", population)

        bio_path = Path(population["biophysical_neuron_models_dir"])
        L.debug("Checking neuron model files: %s", bio_path)
        errors += _check_files(
            f"model_template: {group_name}[{group.file.filename}]",
            (
                bio_path / _get_model_template_file(m)
                for m in group_df.get("model_template", pd.Series(dtype="object")).unique()
            ),
            Error.WARNING,
        )
    else:
        errors.append(
            fatal(f"'biophysical_neuron_models_dir' not defined for population '{population_name}'")
        )
    return errors


def _check_nodes_group(group_df, group, population, population_name):
    """Validates nodes group in nodes population.

    Args:
        group_df (pd.DataFrame): nodes group in nodes .h5 file
        group (h5py.Group): nodes group in nodes .h5 file
        population (dict): the node population config merged with the components config
        population_name (str): the name of the population

    Returns:
        list: List of errors, empty if no errors
    """
    L.debug("Check nodes group: %s", group.name)

    errors = []
    if "model_type" in group_df and group_df["model_type"][0] != population["type"]:
        message = (
            f"Population '{population_name}' type mismatch: "
            f"'{group_df['model_type'][0]}' (nodes_file), "
            f"'{population['type']}' (config)"
        )
        errors.append(Error(Error.WARNING, message))

    if population["type"] == "biophysical":
        return errors + _check_bio_nodes_group(group_df, group, population, population_name)

    return errors


def validate_node_population(nodes_file, population_dict, name):
    """Validates nodes population.

    Args:
        nodes_file (str): path to the nodes file (.h5)
        population_dict (dict): the node population config merged with the components config
        config (dict): resolved bluepysnap config

    Returns:
        list: List of errors, empty if no errors
    """
    with h5py.File(nodes_file, "r") as h5f:
        if "nodes" not in h5f or len(h5f["nodes"]) == 0:
            return []

        # special case in which there are populations but not the one expected
        if name not in h5f["nodes"]:
            return [fatal(f"population '{name}' not found in {nodes_file}")]

        population = h5f[f"nodes/{name}"]

        if "0" in population and "node_type_id" in population:
            group = population["0"]
            group_df = _nodes_group_to_dataframe(group, population)
            if len(group_df) > 0:
                return _check_nodes_group(group_df, group, population_dict, name)

    return []


def _get_node_ids(nodes_h5, population_name):
    """Gets node ids of node population.

    Args:
        nodes_h5: (h5py.File): nodes file h5 instance
        population_name (str): node population name

    Returns:
        np.ndarray: Numpy array of node ids, empty if couldn't find any
    """
    if f"nodes/{population_name}" in nodes_h5:
        node_population = nodes_h5["nodes"][population_name]
        if "node_id" in node_population:
            return node_population["node_id"][:]
        elif "0" in node_population:
            for attr in node_population["0"].values():
                if isinstance(attr, h5py.Dataset):
                    return np.arange(len(attr))

    return np.empty(0)


def _check_edges_node_ids(nodes_ds, nodes):
    """Checks that nodes ids in edges can be resolved to nodes ids in nodes populations.

    Args:
        nodes_ds (h5py.Dataset): nodes dataset in edges population
        nodes (list): "nodes" part of the resolved bluepysnap config

    Returns:
        list: List of errors, empty if no errors
    """
    L.debug("Check edges node ids: %s", nodes_ds.name)

    if "node_population" not in nodes_ds.attrs:
        return []

    node_population_name = nodes_ds.attrs["node_population"]

    nodes_dict = _find_nodes_population(node_population_name, nodes)
    if not nodes_dict:
        return [fatal(f'No node population for "{nodes_ds.name}"')]

    if "nodes_file" not in nodes_dict or not Path(nodes_dict["nodes_file"]).is_file():
        return []

    errors = []
    with h5py.File(nodes_dict["nodes_file"], "r") as h5f:
        node_ids = _get_node_ids(h5f, node_population_name)
        if node_ids.size > 0:
            missing_ids = sorted(set(nodes_ds[:]) - set(node_ids))
            if missing_ids:
                errors.append(
                    fatal(f"{nodes_ds.name} misses node ids in its node population: {missing_ids}")
                )
        elif f"nodes/{node_population_name}" in h5f:
            errors.append(fatal((f"{nodes_ds.name} does not have node ids in its node population")))

    return errors


def _check_edges_indices(population):
    """Check edges population indices.

    Args:
        population (h5py.Group): edges population

    Returns:
        list: List of errors, empty if no errors
    """
    L.debug("Check edges indices: %s", population.name)

    def _check(indices, nodes_ds):
        """The main indices check.

        It iterates over edge indices and verifies that each has its
        nodes in place in nodes populations
        """
        nodes_ranges = indices["node_id_to_ranges"]
        node_to_edges_ranges = indices["range_to_edge_id"]
        for node_id, nodes_range in enumerate(nodes_ranges[:]):
            if 0 <= nodes_range[0] < nodes_range[1]:
                edges_range = node_to_edges_ranges[nodes_range[0] : nodes_range[1]][0]
                edge_node_ids = list(set(nodes_ds[edges_range[0] : edges_range[1]]))
                if len(edge_node_ids) > 1 or edge_node_ids[0] != node_id:
                    errors.append(
                        fatal(
                            f"Population {population.file.filename} edges {edge_node_ids} have "
                            f"node ids {edges_range} instead of single id {node_id}"
                        )
                    )

    errors = []
    indices = population["indices"]
    source_to_target = indices["source_to_target"] if "source_to_target" in indices else None
    target_to_source = indices["target_to_source"] if "target_to_source" in indices else None

    # These are "optional" (not mentioned in our spec) but better to at least give a warning
    if not source_to_target:
        errors.append(Error(Error.WARNING, f'No "source_to_target" in {population.file.filename}'))

    if not target_to_source:
        errors.append(Error(Error.WARNING, f'No "target_to_source" in {population.file.filename}'))

    if target_to_source and source_to_target:
        if "source_node_id" in population:
            _check(source_to_target, population["source_node_id"])
        if "target_node_id" in population:
            _check(target_to_source, population["target_node_id"])

    return errors


def _check_edge_population_data(population, nodes):
    """Check edges population data.

    Args:
        population (h5py.Group): edges population
        nodes (list): "nodes" part of the resolved bluepysnap config

    Returns:
        list: List of errors, empty if no errors
    """
    L.debug("Check edges population data")

    errors = []

    if "source_node_id" in population:
        errors += _check_edges_node_ids(population["source_node_id"], nodes)

    if "target_node_id" in population:
        errors += _check_edges_node_ids(population["target_node_id"], nodes)

    if "indices" in population:
        errors += _check_edges_indices(population)
    else:  # "optional" (not mentioned in our spec) but better to at least give a warning
        errors.append(Error(Error.WARNING, f'No "indices" in {population.file.filename}'))

    return errors


def validate_edge_population(edges_file, name, nodes):
    """Validate an edge population.

    Args:
        edges_file (str): path to the edges file
        name (str): name of the population
        nodes (list): "nodes" listing of the config

    Returns:
        list: List of errors, empty if no errors
    """
    with h5py.File(edges_file, "r") as h5f:
        if "edges" not in h5f or len(h5f["edges"]) == 0:
            return []

        # special case in which there are populations but not the one expected
        if name not in h5f["edges"]:
            return [fatal(f"population '{name}' not found in {edges_file}")]

        population = h5f[f"edges/{name}"]

        if "0" in population:
            return _check_edge_population_data(population, nodes)

    return []


def validate_edges_dict(edges_dict, nodes, skip_slow):
    """Validate an item in the "edges" list.

    Args:
        edges_dict (dict): edges population, represented by an item of "edges" in ``config``
        nodes (list): "nodes" part of the resolved bluepysnap config
        skip_slow(bool): skip slow tests

    Returns:
        list: List of errors, empty if no errors
    """
    errors = []

    def _is_source_node_virtual(edges_dict, edge_population, nodes):
        """Check if source node is virtual.

        The required attributes are different for edges with virtual source nodes.
        """
        with h5py.File(edges_dict["edges_file"], "r") as h5:
            source = h5.get(f"edges/{edge_population}/source_node_id")
            source_population = source.attrs.get("node_population") if source else None

        if source_population:
            nodes_dict = _find_nodes_population(source_population, nodes)
            if nodes_dict is not None:
                return nodes_dict["populations"][source_population].get("type") == "virtual"

        return False

    for name, population in edges_dict.get("populations", {}).items():
        pop_type = population.get("type", DEFAULT_EDGE_TYPE)
        edges_file = edges_dict["edges_file"]

        if Path(edges_file).is_file():
            virtual = False
            if pop_type == "chemical":
                virtual = _is_source_node_virtual(edges_dict, name, nodes)
            errors += schemas.validate_edges_schema(edges_file, pop_type, virtual)
            if not skip_slow:
                errors += validate_edge_population(edges_file, name, nodes)
        else:
            errors.append(fatal(f'Invalid "edges_file": {edges_file}'))

    return errors


def validate_nodes_dict(nodes_dict, components):
    """Validate an item in the "nodes" list.

    Args:
        nodes_dict (dict): nodes population, represented by an item of "nodes" in ``config``
        components(dict): "components" part of the ``config``

    Returns:
        list: List of errors, empty if no errors
    """
    errors = []
    for pop_name, pop_dict in nodes_dict.get("populations", {}).items():
        population = {**components, **pop_dict}
        population["type"] = population.get("type", DEFAULT_NODE_TYPE)
        nodes_file = nodes_dict["nodes_file"]

        if Path(nodes_file).is_file():
            errors = schemas.validate_nodes_schema(nodes_file, population["type"])
            errors += validate_node_population(nodes_file, population, pop_name)
        else:
            errors.append(fatal(f'Invalid "nodes_file": {nodes_file}'))

    return errors


def validate_networks(config, skip_slow):
    """Validate "networks" part of the config.

    Acts as a starting point of validation.
    """
    errors = []
    errors += _check_duplicate_populations(config["networks"], "nodes")
    errors += _check_duplicate_populations(config["networks"], "edges")

    components = config.get("components", {})
    nodes = config["networks"].get("nodes", [])

    for nodes_dict in nodes:
        if "nodes_file" in nodes_dict:
            errors += validate_nodes_dict(nodes_dict, components)
    for edges_dict in config["networks"].get("edges", []):
        if "edges_file" in edges_dict:
            errors += validate_edges_dict(edges_dict, nodes, skip_slow)

    return errors


def validate(config_file, skip_slow, only_errors=False, print_errors=True):
    """Validates Sonata circuit.

    Args:
        config_file (str): path to Sonata circuit config file official checks.
        skip_slow (bool): skip slow tests
        only_errors (bool): only return/print fatal errors
        print_errors (bool): print errors

    Returns:
        list: List of errors, empty if no errors
    """
    config = Parser.parse(load_json(config_file), str(Path(config_file).parent))
    errors = schemas.validate_circuit_schema(config_file, config)

    if "networks" in config:
        errors += validate_networks(config, skip_slow)

    if _check_partial_circuit_config(config):
        message = (
            "The Circuit config is partial. Validity cannot be established "
            "for partial configs as it depends on the intended use. "
        )
        L.warning(message)
        errors.append(Error(Error.WARNING, message))

    if only_errors:
        errors = [e for e in errors if e.level == Error.FATAL]

    if print_errors:
        _print_errors(errors)

    return set(errors)



import sys
from unittest.mock import patch

import pytest

import bluepysnap._plotting as test_module


def test__get_pyplot():
    with patch.dict(sys.modules, {"matplotlib.pyplot": None}):
        with pytest.raises(ImportError):
            test_module._get_pyplot()

    import matplotlib.pyplot

    plt_test = test_module._get_pyplot()
    assert plt_test is matplotlib.pyplot



import json
import os
import pickle
import warnings

import libsonata
import pytest
from libsonata import SonataError

import bluepysnap.simulation as test_module
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.frame_report import (
    CompartmentReport,
    PopulationCompartmentReport,
    PopulationSomaReport,
    SomaReport,
)
from bluepysnap.node_sets import NodeSets
from bluepysnap.spike_report import PopulationSpikeReport, SpikeReport

from utils import PICKLED_SIZE_ADJUSTMENT, TEST_DATA_DIR, copy_test_data, edit_config


def test__warn_on_overwritten_node_sets():
    # No warnings if no overwritten
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        test_module._warn_on_overwritten_node_sets(set())

    with pytest.warns(RuntimeWarning, match="Simulation node sets overwrite 3 .*: a, b, c"):
        test_module._warn_on_overwritten_node_sets(["a", "b", "c"])

    with pytest.warns(RuntimeWarning, match=r"Simulation node sets overwrite 3 .*: a, b, \.\.\."):
        test_module._warn_on_overwritten_node_sets(["a", "b", "c"], print_max=2)


def test_all():
    simulation = test_module.Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
    assert simulation.config["network"] == str(TEST_DATA_DIR / "circuit_config.json")
    assert set(simulation.circuit.nodes) == {"default", "default2"}
    assert set(simulation.circuit.edges) == {"default", "default2"}

    assert isinstance(simulation.run, libsonata._libsonata.Run)
    assert simulation.run.tstop == 1000.0
    assert simulation.run.dt == 0.01
    assert simulation.run.spike_threshold == -15
    assert simulation.run.random_seed == 42
    assert simulation.time_start == 0.0
    assert simulation.time_stop == 1000.0
    assert simulation.dt == 0.01
    assert simulation.time_units == "ms"

    assert simulation.simulator == "CORENEURON"
    assert isinstance(simulation.conditions, libsonata._libsonata.Conditions)
    assert simulation.conditions.celsius == 34.0
    assert simulation.conditions.v_init == -80

    with pytest.warns(RuntimeWarning, match="Simulation node sets overwrite 1 .* Layer23"):
        assert isinstance(simulation.node_sets, NodeSets)

    expected_content = {
        **json.loads((TEST_DATA_DIR / "node_sets.json").read_text()),
        "Layer23": {"layer": [2, 3]},
    }
    assert simulation.node_sets.content == expected_content

    assert isinstance(simulation.inputs, dict)
    assert isinstance(simulation.spikes, SpikeReport)
    assert isinstance(simulation.spikes["default"], PopulationSpikeReport)

    assert set(simulation.reports) == {"soma_report", "section_report"}
    assert isinstance(simulation.reports["soma_report"], SomaReport)
    assert isinstance(simulation.reports["section_report"], CompartmentReport)

    rep = simulation.reports["soma_report"]
    assert set(rep.population_names) == {"default", "default2"}
    assert isinstance(rep["default"], PopulationSomaReport)

    rep = simulation.reports["section_report"]
    assert set(rep.population_names) == {"default", "default2"}
    assert isinstance(rep["default"], PopulationCompartmentReport)


def test_no_warning_when_shared_node_sets_path():
    with copy_test_data(config="simulation_config.json") as (_, config_path):
        circuit = test_module.Simulation(config_path).circuit
        circuit_node_sets_path = circuit.to_libsonata.node_sets_path

        # set simulation node set path = circuit node set path
        with edit_config(config_path) as config:
            config["node_sets_file"] = circuit_node_sets_path

        # Should not raise
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            test_module.Simulation(config_path).node_sets

        # if node_sets_file is not defined, libsonata should use the same path as the circuit
        with edit_config(config_path) as config:
            config.pop("node_sets_file")

        # Should not raise either
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            test_module.Simulation(config_path).node_sets


def test_unknown_report():
    with copy_test_data(config="simulation_config.json") as (_, config_path):
        with edit_config(config_path) as config:
            config["reports"]["soma_report"]["sections"] = "unknown"

        with pytest.raises(libsonata.SonataError, match="Invalid value.*for key 'sections'"):
            test_module.Simulation(config_path)


def test_nonimplemented_report():
    with copy_test_data(config="simulation_config.json") as (_, config_path):
        with edit_config(config_path) as config:
            config["reports"]["soma_report"]["sections"] = "axon"

        with pytest.raises(
            BluepySnapError, match="Report soma_report: format axon not yet supported"
        ):
            test_module.Simulation(config_path).reports


def test_network_file_not_found():
    with copy_test_data(config="simulation_config.json") as (_, config_path):
        with edit_config(config_path) as config:
            config.pop("network")

        os.remove(config_path.parent / "circuit_config.json")
        simulation = test_module.Simulation(config_path)

        with pytest.raises(BluepySnapError, match="'network' file not found"):
            simulation.circuit


def test_no_node_set():
    with copy_test_data(config="simulation_config.json") as (_, config_path):
        with edit_config(config_path) as config:
            config.pop("node_sets_file")
            # remove circuit config to prevent libsonata from fetching the path from there
            os.remove(config_path.parent / "circuit_config.json")

        simulation = test_module.Simulation(config_path)
        assert simulation.node_sets.content == {}


def test_pickle(tmp_path):
    pickle_path = tmp_path / "pickle.pkl"
    simulation = test_module.Simulation(TEST_DATA_DIR / "simulation_config.json")

    # trigger some cached properties, to makes sure they aren't being pickeld
    simulation.circuit

    with open(pickle_path, "wb") as fd:
        pickle.dump(simulation, fd)

    with open(pickle_path, "rb") as fd:
        simulation = pickle.load(fd)

    assert pickle_path.stat().st_size < 70 + PICKLED_SIZE_ADJUSTMENT
    assert simulation.dt == 0.01



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Configuration variables."""

import os

# All possible checks enabled / deprecated methods disallowed
STRICT_MODE = False


def str2bool(value):
    """Convert environment variable value to bool."""
    if value is None:
        return False
    else:
        return value.lower() in ("y", "yes", "true", "1")


def load_env():
    """Load settings from environment variables."""
    # pylint: disable=global-statement
    if "BLUESNAP_STRICT_MODE" in os.environ:
        global STRICT_MODE
        STRICT_MODE = str2bool(os.environ["BLUESNAP_STRICT_MODE"])


load_env()



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Plotting module for the different snap objects."""
import logging

import numpy as np
import pandas as pd
from more_itertools import roundrobin

from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Node
from bluepysnap.utils import IDS_DTYPE

L = logging.getLogger(__name__)


def _get_pyplot():
    try:
        import matplotlib.pyplot as plt
    except ImportError as e:
        raise ImportError(
            "Bluepysnap requirements are not installed.\n"
            "Please pip install as follows:\n"
            "  pip install bluepysnap[plots] --upgrade"
        ) from e
    return plt


def spikes_firing_rate_histogram(filtered_report, time_binsize=None, ax=None):  # pragma: no cover
    """Spike firing rate histogram.

    This plot shows the number of nodes firing during a range of time.

    Args:
        time_binsize(None/int/float): bin size (milliseconds). If None, a binning heuristic is used
            to create an histogram with ~100 spikes per bin in average.
        ax(matplotlib.Axis): matplotlib Axis to draw on (if not specified, pyplot.gca() is used).

    Returns:
        matplotlib.Axis: Axis containing firing rate histogram.

    Notes:
        If no axis is provided through the ax=ax keyword argument,
        then a default layout is set using pyplot.gca().
    """
    # pylint: disable=too-many-locals
    plt = _get_pyplot()
    if time_binsize is not None and time_binsize <= 0:
        raise BluepySnapError(f"Invalid time_binsize = {time_binsize}. Should be > 0.")

    if ax is None:
        ax = plt.gca()
        ax.set_xlabel("Time [ms]")
        ax.set_ylabel("PSTH [Hz]")

    times = filtered_report.report.index

    if len(times) > 0:
        time_start = np.min(times)
        time_stop = np.max(times)

        if time_binsize is None:
            # heuristic for a nice bin size (~100 spikes per bin on average)
            time_binsize = min(50.0, (time_stop - time_start) / ((len(times) / 100.0) + 1.0))

        bins = np.append(np.arange(time_start, time_stop, time_binsize), time_stop)
        hist, bin_edges = np.histogram(times, bins=bins)
        node_count = filtered_report.report[["ids", "population"]].drop_duplicates().shape[0]
        freq = 1.0 * hist / node_count / (0.001 * time_binsize)

        # use the middle of the bins instead of the start of the bin
        ax.plot(0.5 * (bin_edges[1:] + bin_edges[:-1]), freq, label="PSTH", drawstyle="steps-mid")

    return ax


def spike_raster(filtered_report, y_axis=None, ax=None):  # pragma: no cover
    """Spike raster plot.

    Shows a global overview of the circuit's firing nodes. The y axis can project either the
    node_ids or any properties present in the different node populations.

    Args:
        y_axis (None/str): The property to display on the y axis. None is node_ids.
        ax(matplotlib.Axis): matplotlib Axis to draw on (if not specified, pyplot.gca() is used).

    Returns:
        matplotlib.Axis: Axis containing Spikes raster plot.

    Notes:
        If no axis is provided through the ax=ax keyword argument,
        then a default layout is set using pyplot.gca().
    """
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    plt = _get_pyplot()

    spike_report = filtered_report.spike_report
    population_names = filtered_report.spike_report.population_names

    props = {
        "node_id_offset": 0,
        "pop_separators": [],
        "categorical_values": set(),
        "ymin": np.inf,
        "ymax": -np.inf,
    }

    def _update_raster_properties():
        if y_axis is None:
            props["node_id_offset"] += spikes.nodes.size
            props["pop_separators"].append(props["node_id_offset"])
        elif isinstance(spikes.nodes.property_dtypes[y_axis], pd.CategoricalDtype):
            props["categorical_values"].update(spikes.nodes.property_values(y_axis))
        else:
            props["ymin"] = min(props["ymin"], spikes.nodes.get(properties=y_axis).min())
            props["ymax"] = max(props["ymax"], spikes.nodes.get(properties=y_axis).max())

    report = filtered_report.report

    # use np.int64 if displaying node_ids
    dtype = spike_report[population_names[0]].nodes.property_dtypes[y_axis] if y_axis else IDS_DTYPE
    if isinstance(dtype, pd.CategoricalDtype):
        # this is to prevent the problems when concatenating categoricals with unknown categories
        dtype = str
    data = pd.Series(index=report.index, dtype=dtype)
    for population in population_names:
        spikes = spike_report[population]
        mask = report["population"] == population
        if y_axis is None:
            data.loc[mask] = report.loc[mask, "ids"] + props["node_id_offset"]
        else:
            ids = report.loc[mask, "ids"].to_numpy()
            try:
                ys = spikes.nodes.get(properties=y_axis)
            except BluepySnapError:
                continue
            # astype is used to avoid problems with the categorical
            data[mask] = ys[ids].astype(dtype).to_numpy()
        _update_raster_properties()

    data = data[data.notna()]
    if ax is None:
        ax = plt.gca()
        ax.xaxis.grid()
        ax.set_xlabel("Time [ms]")
        ax.tick_params(axis="y", which="both", length=0)
        ax.set_xlim(spike_report.time_start, spike_report.time_stop)
        if y_axis is None:
            ax.set_ylim(0, props["node_id_offset"])
            ax.set_ylabel("nodes")
        else:
            if np.issubdtype(type(data.iloc[0]), np.number):
                # automatically expended by plt if ymin == ymax
                ax.set_ylim(props["ymin"], props["ymax"])
            else:
                labels = sorted(list(props["categorical_values"]))
                ax.set_yticks(np.arange(len(labels)))
                ax.set_yticklabels(labels)
                if len(labels) > 1:
                    ax.set_ylim(-0.5, len(labels) - 0.5)
            ax.set_ylabel(y_axis)

    ax.scatter(data.index.to_numpy(), data.to_numpy(), s=10, marker="|")
    if len(props["pop_separators"]) > 1:
        for separator in props["pop_separators"]:
            ax.axhline(y=separator, color="black", lw=2)
    return ax


def spikes_isi(filtered_report, use_frequency=False, binsize=None, ax=None):  # pragma: no cover
    # pylint: disable=too-many-locals
    """Interspike interval histogram.

    This plots show the binned time/frequency interval between to spikes for neurons.

    Args:
        use_frequency(bool): use inverse interspike interval times (Hz)
        binsize(None/int/float): bin size in milliseconds or Hz. If None is used the binning is
            delegated to matplolib and is done automatically.
        ax(matplotlib.Axis): matplotlib Axis to draw on (if not specified, pyplot.gca() is used).

    Returns:
        matplotlib.Axis: axis containing the interspike interval histogram.

    Notes:
        If no axis is provided through the ax=ax keyword argument,
        then a default layout is set using pyplot.gca().
    """
    plt = _get_pyplot()
    if binsize is not None and binsize <= 0:
        raise BluepySnapError(f"Invalid binsize = {binsize}. Should be > 0.")

    gb = filtered_report.report.groupby(["ids", "population"])
    values = np.concatenate([np.diff(node_spikes.index.to_numpy()) for _, node_spikes in gb])

    if len(values) == 0:
        raise BluepySnapError(
            f"No data to display. You should check your 'group' query: {filtered_report.group}."
        )
    if use_frequency:
        values = values[values > 0]  # filter out zero intervals
        values = 1000.0 / values

    if binsize is None:
        bins = "auto"
    else:
        bins = np.arange(0, np.max(values), binsize)

    if ax is None:
        ax = plt.gca()
        if use_frequency:
            ax.set_xlabel("Frequency [Hz]")
        else:
            ax.set_xlabel("Interspike interval [ms]")
        ax.set_ylabel("Bin weight")

    ax.hist(values, bins=bins, edgecolor="black", density=True)
    return ax


def spikes_firing_animation(
    filtered_report, x_axis=Node.X, y_axis=Node.Y, dt=20, ax=None
):  # pragma: no cover
    # pylint: disable=too-many-locals,too-many-arguments,anomalous-backslash-in-string
    """Simple animation of simulation spikes.

    Each frame of the animation represents the spiking nodes during a period of dt ms seconds
    in a coordinate system corresponding to the x, y or z axis of the circuit.

    Args:
        x_axis (str): Node enum that will determine the animation x_axis
        y_axis (str): Node enum that will determine the animation y_axis
        dt (int) : the time bin size of each frame in the video in ms
        ax(matplotlib.Axis): matplotlib Axis to draw on (if not specified, pyplot.gca()
            and plt.figure() are used).

    Returns:
        (matplotlib.animation.FuncAnimation, matplotlib.Axis): the matplotlib animation object and
        the corresponding axis.

    Examples:
        From scripts:

        >>> import matplotlib.pyplot as plt
        >>> from bluepysnap import Simulation
        >>> report = Simulation("config.json").spikes["my_population"]
        >>> anim, ax = report.firing_animation()
        >>> plt.show()
        >>> # to save the animation : do not plt.show() and just anim.save('my_movie.mp4')

        From notebooks:

        >>> from IPython.display import HTML
        >>> from bluepysnap import Simulation
        >>> report = Simulation("config.json").spikes["my_population"]
        >>> anim, ax = report.firing_animation()
        >>> HTML(anim.to_html5_video())
    """
    plt = _get_pyplot()
    from matplotlib.animation import FuncAnimation

    def _check_axis(axis):
        """Verifies axes values."""
        axes = {Node.X, Node.Y, Node.Z}
        if axis not in axes:
            raise BluepySnapError(f"{axis} is not a valid axis")

    _check_axis(x_axis)
    _check_axis(y_axis)

    spike_report = filtered_report.spike_report
    population_names = filtered_report.spike_report.population_names
    report = filtered_report.report

    data = pd.DataFrame(index=report.index, columns=[x_axis, y_axis], dtype=np.float32)

    for population in population_names:
        spikes = spike_report[population]
        pop_mask = report["population"] == population

        ids = report.loc[pop_mask, "ids"].to_numpy()
        try:
            values = spikes.nodes.get(properties=[x_axis, y_axis]).loc[ids].to_numpy()
            data.loc[pop_mask, [x_axis, y_axis]] = values
        except BluepySnapError:
            continue

    data = data[data.notna()]

    if ax is None:
        fig = plt.figure()
        ax = plt.gca()
        ax.set_title(f"time = {np.min(data.index)}ms")
        x_limits = [data[x_axis].min(), data[x_axis].max()]
        y_limits = [data[y_axis].min(), data[y_axis].max()]
        ax.set_xlim(*x_limits)
        ax.set_ylim(*y_limits)
        ax.set_xlabel(rf"{x_axis} $\mu$m")  # noqa
        ax.set_ylabel(rf"{y_axis} $\mu$m")  # noqa

    else:
        fig = ax.figure

    dots = ax.plot([], [], ".k")

    def update_animation(frame):
        """Update the animation plots and axes."""
        ax.set_title("time = " + str(frame * dt) + " ms")
        mask = (data.index >= frame * dt) & (data.index <= (frame + 1) * dt)
        positions = data.loc[mask, [x_axis, y_axis]].values
        x = positions[:, 0]
        y = positions[:, 1]
        dots[0].set_data(x, y)
        return dots

    frames = list(range(int(data.index[0] / dt), int(data.index[-1] / dt)))
    anim = FuncAnimation(fig, update_animation, frames=frames)
    return anim, ax


def frame_trace(filtered_report, plot_type="mean", ax=None):  # pragma: no cover
    """Returns a plot displaying the voltage of a node or a compartment as a function of time.

    Args:
        plot_type (str): string either `all` or `mean`. `all` will plot the first 15 traces from the
            group. `mean` will plot the mean value of the node
        ax: A plot axis object that will be updated

    Returns:
        matplotlib.Axis: axis containing the soma's traces.
    """
    # pylint: disable=too-many-locals

    plt = _get_pyplot()

    if ax is None:
        ax = plt.gca()
        data_units = filtered_report.frame_report.data_units
        if plot_type == "mean":
            ax.set_ylabel(f"Avg volt. [{data_units}]")
        elif plot_type == "all":
            ax.set_ylabel(f"Voltage [{data_units}]")
        ax.set_xlabel(f"Time [{filtered_report.frame_report.time_units}]")
        ax.set_xlim([filtered_report.report.index.min(), filtered_report.report.index.max()])

    if plot_type == "mean":
        ax.plot(filtered_report.report.T.mean())
    elif plot_type == "all":
        max_per_pop = 15
        levels = filtered_report.report.columns.levels
        slicer = []
        # create a slicer that will slice only on the last level of the columns
        # that is, node_id for the soma report, element_id for the compartment report
        for i, level in enumerate(levels):
            max_ = level[:max_per_pop][-1]
            slicer.append(slice(None) if i != len(levels) - 1 else slice(None, max_))
        data = filtered_report.report.loc[:, tuple(slicer)].T
        # create [[(pop1, id1), (pop1, id2),...], [(pop2, id1), (pop2, id2),...]]
        indexes = [[(pop, idx) for idx in data.loc[pop].index] for pop in levels[0]]
        # try to keep the maximum of ids from each population
        kept_ids = list(roundrobin(*indexes))[:max_per_pop]
        for _, row in data.loc[kept_ids].iterrows():
            ax.plot(row)
    else:
        raise BluepySnapError(f"Unknown plot_type {plot_type}. Should be 'mean or 'all'.")
    return ax



#!/usr/bin/env python

# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from setuptools import find_packages, setup
from setuptools.command.egg_info import egg_info


# setuptools_scm forcibly includes all files under version control into the sdist
# See https://github.com/pypa/setuptools_scm/issues/190
# Workaround taken from:
# https://github.com/raiden-network/raiden/commit/3fb837e8b6e6343f65d99055459cb440e1a938ff
class EggInfo(egg_info):
    def __init__(self, *args, **kwargs):
        egg_info.__init__(self, *args, **kwargs)
        try:
            import setuptools_scm.integration

            setuptools_scm.integration.find_files = lambda _: []
        except ImportError:
            pass


with open("README.rst") as f:
    README = f.read()

setup(
    name="bluepysnap",
    python_requires=">=3.8",
    install_requires=[
        "cached_property>=1.0",
        "h5py>=3.0.1,<4.0.0",
        "importlib_resources>=5.0.0",
        "jsonschema>=4.18.0,<5.0.0",
        "referencing>=0.28.4",
        "libsonata>=0.1.24,<1.0.0",
        "morphio>=3.0.0,<4.0.0",
        "morph-tool>=2.4.3,<3.0.0",
        "numpy>=1.8",
        "pandas>=1.0.0",
        "click>=7.0",
        "more-itertools>=8.2.0",
    ],
    extras_require={
        "docs": ["sphinx", "sphinx-bluebrain-theme"],
        "plots": ["matplotlib>=3.0.0"],
        "spatial-index": ["spatial-index>=1.2.1"],
    },
    packages=find_packages(),
    package_data={
        "bluepysnap": ["schemas/*.yaml", "schemas/*/*.yaml"],
    },
    use_scm_version={
        "write_to": "bluepysnap/_version.py",
    },
    setup_requires=[
        "setuptools_scm",
    ],
    cmdclass={
        "egg_info": EggInfo,
    },
    entry_points="""
        [console_scripts]
        bluepysnap=bluepysnap.cli:cli
    """,
    author="Blue Brain Project, EPFL",
    description="Simulation and Neural network Analysis Productivity layer",
    long_description=README,
    long_description_content_type="text/x-rst",
    license="LGPLv3",
    url="https://github.com/BlueBrain/snap",
    keywords=["SONATA", "BlueBrainProject"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX",
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
)



from unittest.mock import Mock, patch

import h5py
import numpy as np
import pytest

import bluepysnap.circuit_validation as test_module
from bluepysnap.circuit_validation import Error
from bluepysnap.exceptions import BluepySnapError

from utils import TEST_DATA_DIR, copy_test_data, edit_config


@patch("bluepysnap.schemas.validate_nodes_schema")
@patch("bluepysnap.schemas.validate_edges_schema")
@patch("bluepysnap.schemas.validate_circuit_schema")
def validate(s, *mocks, print_errors=False):
    # reset the mock return value to empty list
    for m in mocks:
        m.return_value = []
    return test_module.validate(s, skip_slow=False, print_errors=print_errors)


def test_error_comparison():
    err = Error(Error.WARNING, "hello")
    assert err != "hello"


def test_ok_circuit():
    errors = validate(str(TEST_DATA_DIR / "circuit_config.json"))
    assert errors == set()

    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["node_types_file"] = None
        errors = validate(str(config_copy_path))
        assert errors == set()


@patch("bluepysnap.schemas.validate_nodes_schema", Mock(return_value=[]))
@patch("bluepysnap.schemas.validate_edges_schema", Mock(return_value=[]))
@patch("bluepysnap.schemas.validate_circuit_schema", Mock(return_value=[]))
def test_skip_slow():
    with patch(f"{test_module.__name__}.validate_edge_population") as patched:
        errors = test_module.validate(
            str(TEST_DATA_DIR / "circuit_config.json"), skip_slow=True, print_errors=False
        )
        assert errors == set()
        patched.assert_not_called()


def test_only_errors():
    with patch(f"{test_module.__name__}.validate_networks") as patched:
        patched.return_value = [
            Error(Error.FATAL, "fatal_error"),
            Error(Error.WARNING, "a mere warning"),
            Error(Error.INFO, "utterly useful message"),
        ]

        errors = test_module.validate(
            str(TEST_DATA_DIR / "circuit_config.json"),
            skip_slow=True,
            only_errors=True,
        )
        assert len(errors) == 1
        assert list(errors)[0] == Error(Error.FATAL, "fatal_error")


def test_print_errors(capsys):
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            del config["networks"]["nodes"]
        validate(str(config_copy_path), print_errors=True)
    assert "No node population for" in capsys.readouterr().out


@pytest.mark.parametrize(
    "to_remove",
    (
        ["networks"],
        ["networks", "nodes", 0, "nodes_file"],
        ["networks", "edges"],
        ["networks", "edges", 0, "edges_file"],
        ["networks", "edges", 0, "populations"],
        ["networks", "edges", 0, "populations", "default"],
    ),
)
def test_missing_data_config_no_error(to_remove):
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            c = config
            for key in to_remove[:-1]:
                c = c[key]
            del c[to_remove[-1]]
        errors = validate(str(config_copy_path))
        assert errors == set()


@pytest.mark.parametrize(
    "to_remove",
    (
        ["networks", "nodes"],
        ["networks", "nodes", 0, "populations"],
        ["networks", "nodes", 0, "populations", "default"],
    ),
)
def test_missing_data_config_no_population_for_edge(to_remove):
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            c = config
            for key in to_remove[:-1]:
                c = c[key]
            del c[to_remove[-1]]
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(Error.FATAL, 'No node population for "/edges/default/source_node_id"'),
            Error(Error.FATAL, 'No node population for "/edges/default/target_node_id"'),
            Error(Error.FATAL, 'No node population for "/edges/default2/source_node_id"'),
            Error(Error.FATAL, 'No node population for "/edges/default2/target_node_id"'),
        }


def test_nodes_population_not_found_in_h5():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["fake_population"] = {}
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.FATAL,
                f"population 'fake_population' not found in {nodes_file}",
            )
        }


def test_edges_population_not_found_in_h5():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with edit_config(config_copy_path) as config:
            config["networks"]["edges"][0]["populations"]["fake_population"] = {}
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.FATAL,
                f"population 'fake_population' not found in {edges_file}",
            )
        }


def test_ok_node_population_type():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["default"]["type"] = "biophysical"
        errors = validate(str(config_copy_path))
        assert errors == set()


@pytest.mark.parametrize("model_type", ("virtual", "electrical", "fake_type"))
def test_ok_nonbio_type(model_type):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["default"]["type"] = model_type
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            h5f["nodes/default/0/model_type"][:] = model_type
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_population_type_mismatch():
    with copy_test_data() as (_, config_copy_path):
        fake_type = "fake_type"
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["default"]["type"] = fake_type
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                (
                    "Population 'default' type mismatch: "
                    f"'biophysical' (nodes_file), '{fake_type}' (config)"
                ),
            )
        }


def test_ok_edge_population_type():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["edges"][0]["populations"]["default"]["type"] = "chemical"
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_invalid_edge_population_type():
    with copy_test_data() as (_, config_copy_path):
        fake_type = "fake_type"
        with edit_config(config_copy_path) as config:
            config["networks"]["edges"][0]["populations"]["default"]["type"] = fake_type
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_invalid_config_nodes_file():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["nodes_file"] = "/"
        errors = validate(str(config_copy_path))
        assert errors == {Error(Error.FATAL, 'Invalid "nodes_file": /')}


def test_invalid_config_edges_file():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["edges"][0]["edges_file"] = "/"
        errors = validate(str(config_copy_path))
        assert errors == {Error(Error.FATAL, 'Invalid "edges_file": /')}


@pytest.mark.parametrize(
    "to_remove",
    (
        "nodes",
        "nodes/default/0/mtype",
        "nodes/default/0/morphology",
        "nodes/default/0/model_template",
        "nodes/default/0/model_type",
        "nodes/default/0/dynamics_params",
        "nodes/default/0/layer",
        "nodes/default/0/x",
        "nodes/default/0/y",
        "nodes/default/0/z",
        "nodes/default/0/rotation_angle_xaxis",
        "nodes/default/0/rotation_angle_yaxis",
        "nodes/default/0/rotation_angle_zaxis",
    ),
)
def test_missing_data_nodes_h5_no_error(to_remove):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            del h5f[to_remove]
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_ok_node_ids_dataset():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            h5f["nodes/default/node_id"] = list(range(len(h5f["nodes/default/node_type_id"])))
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_ok_bio_model_type_in_library():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            data = h5f["nodes/default/0/model_type"][:]
            del h5f["nodes/default/0/model_type"]
            h5f.create_dataset("nodes/default/0/model_type", data=np.zeros_like(data, dtype=int))
            h5f.create_dataset(
                "nodes/default/0/@library/model_type",
                data=np.array(
                    [
                        "biophysical",
                    ],
                    dtype=h5py.string_dtype(),
                ),
            )
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_no_bio_component_dirs():
    dirs = ["morphologies_dir", "biophysical_neuron_models_dir"]
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            for dir_ in dirs:
                del config["components"][dir_]
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.FATAL, "'biophysical_neuron_models_dir' not defined for population 'default'"
            ),
            Error(
                Error.FATAL,
                (
                    "at least one of 'morphologies_dir' or 'alternate_morphologies' "
                    "must to be defined for 'biophysical' population 'default'"
                ),
            ),
            Error(
                Error.FATAL, "'biophysical_neuron_models_dir' not defined for population 'default2'"
            ),
            Error(
                Error.FATAL,
                (
                    "at least one of 'morphologies_dir' or 'alternate_morphologies' "
                    "must to be defined for 'biophysical' population 'default2'"
                ),
            ),
        }


def test_invalid_bio_alternate_morphology_dir():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        component = "neurolucida-asc"
        fake_path = str(circuit_copy_path / "fake/path")
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["default"]["alternate_morphologies"] = {
                component: fake_path
            }
        errors = validate(str(config_copy_path))
        assert errors == {Error(Error.FATAL, f'Invalid components "{component}": {fake_path}')}


@patch("bluepysnap.circuit_validation.MAX_MISSING_FILES_DISPLAY", 1)
def test_no_morph_files():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            h5f["nodes/default/0/morphology"][0] = "noname"
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                f"missing at least 1 files in group morphology: default/0[{nodes_file}]:\n\tnoname.swc\n",
            )
        }


def test_no_alternate_morph_files():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = str(circuit_copy_path / "nodes.h5")
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"][0]["populations"]["default"]["alternate_morphologies"] = {
                "neurolucida-asc": config["components"]["morphologies_dir"]
            }
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                f"missing at least 1 files in group morphology: default/0[{nodes_file}]:\n\tmorph-A.asc\n",
            )
        }


@patch("bluepysnap.circuit_validation.MAX_MISSING_FILES_DISPLAY", 1)
def test_no_morph_library_files():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            grp = h5f["nodes/default/0"]
            str_dtype = h5py.special_dtype(vlen=str)
            grp.create_dataset("@library/morphology", shape=(1,), dtype=str_dtype)
            grp["@library/morphology"][:] = "noname"
            shape = grp["morphology"].shape
            del grp["morphology"]
            grp.create_dataset("morphology", shape=shape, fillvalue=0, dtype=int)
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                f"missing at least 1 files in group morphology: default/0[{nodes_file}]:\n\tnoname.swc\n",
            )
        }


def test_no_template_files():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            h5f["nodes/default/0/model_template"][0] = "hoc:noname"
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                f"missing at least 1 files in group model_template: default/0[{nodes_file}]:\n\tnoname.hoc\n",
            )
        }


@patch("bluepysnap.circuit_validation.MAX_MISSING_FILES_DISPLAY", 1)
def test_no_template_library_files():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            grp = h5f["nodes/default/0"]
            str_dtype = h5py.special_dtype(vlen=str)
            grp.create_dataset("@library/model_template", shape=(1,), dtype=str_dtype)
            grp["@library/model_template"][:] = "hoc:noname"
            shape = grp["model_template"].shape
            del grp["model_template"]
            grp.create_dataset("model_template", shape=shape, fillvalue=0, dtype=int)
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                f"missing at least 1 files in group model_template: default/0[{nodes_file}]:\n\tnoname.hoc\n",
            )
        }


@pytest.mark.parametrize(
    "to_remove",
    (
        "/edges",
        "/edges/default/edge_type_id",
        "/edges/default/source_node_id",
        "/edges/default/target_node_id",
        "/edges/default/0",
        "/edges/default/0/afferent_center_x",
        "/edges/default/0/afferent_center_y",
        "/edges/default/0/afferent_center_z",
        "/edges/default/0/afferent_section_id",
        "/edges/default/0/afferent_section_pos",
        "/edges/default/0/afferent_surface_x",
        "/edges/default/0/afferent_surface_y",
        "/edges/default/0/afferent_surface_z",
        "/edges/default/0/conductance",
        "/edges/default/0/delay",
        "/edges/default/0/dynamics_params",
        "/edges/default/0/dynamics_params/param1",
        "/edges/default/0/efferent_center_x",
        "/edges/default/0/efferent_center_y",
        "/edges/default/0/efferent_center_z",
        "/edges/default/0/efferent_section_id",
        "/edges/default/0/efferent_section_pos",
        "/edges/default/0/efferent_surface_x",
        "/edges/default/0/efferent_surface_y",
        "/edges/default/0/efferent_surface_z",
        "/edges/default/0/syn_weight",
    ),
)
def test_missing_data_edges_h5_no_error(to_remove):
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f[to_remove]
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_no_edge_indices():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default/indices"]
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(Error.WARNING, f'No "indices" in {edges_file}'),
        }


def test_no_edge_source_to_target():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default/indices/source_to_target"]
            del h5f["edges/default/indices/target_to_source"]
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(Error.WARNING, f'No "source_to_target" in {edges_file}'),
            Error(Error.WARNING, f'No "target_to_source" in {edges_file}'),
        }


def test_no_edge_all_node_ids():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        nodes_file = circuit_copy_path / "nodes.h5"
        with h5py.File(nodes_file, "r+") as h5f:
            del h5f["nodes/default/0"]
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.FATAL,
                "/edges/default/source_node_id does not have node ids in its node population",
            ),
            Error(
                Error.FATAL,
                "/edges/default/target_node_id does not have node ids in its node population",
            ),
            Error(
                Error.FATAL,
                "/edges/default2/source_node_id does not have node ids in its node population",
            ),
            Error(
                Error.FATAL,
                "/edges/default2/target_node_id does not have node ids in its node population",
            ),
        }


def test_invalid_edge_node_ids():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with h5py.File(edges_file, "r+") as h5f:
            h5f["edges/default/source_node_id"][0] = 99999
            h5f["edges/default/target_node_id"][0] = 99999
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.FATAL,
                "/edges/default/source_node_id misses node ids in its node population: [99999]",
            ),
            Error(
                Error.FATAL,
                "/edges/default/target_node_id misses node ids in its node population: [99999]",
            ),
            Error(
                Error.FATAL,
                f"Population {edges_file} edges [99999] have node ids [0 1] instead of "
                "single id 2",
            ),
            Error(
                Error.FATAL,
                f"Population {edges_file} edges [99999] have node ids [0 1] instead of "
                "single id 0",
            ),
        }


def test_explicit_edges_no_node_population_attr():
    with copy_test_data() as (circuit_copy_path, config_copy_path):
        edges_file = circuit_copy_path / "edges.h5"
        with h5py.File(edges_file, "r+") as h5f:
            del h5f["edges/default/source_node_id"].attrs["node_population"]
        errors = validate(str(config_copy_path))
        assert errors == set()


def test_no_duplicate_population_names():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["networks"]["nodes"].append(config["networks"]["nodes"][0])
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(Error.FATAL, 'Already have population "default" in config for type "nodes"'),
            Error(Error.FATAL, 'Already have population "default2" in config for type "nodes"'),
        }


def test_partial_config_warning():
    with copy_test_data() as (_, config_copy_path):
        with edit_config(config_copy_path) as config:
            config["metadata"] = {"status": "partial"}
        errors = validate(str(config_copy_path))
        assert errors == {
            Error(
                Error.WARNING,
                (
                    "The Circuit config is partial. Validity cannot be established "
                    "for partial configs as it depends on the intended use. "
                ),
            ),
        }



import h5py
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

import bluepysnap.morph as test_module
from bluepysnap.circuit import Circuit
from bluepysnap.circuit_ids import CircuitNodeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Node

from utils import TEST_DATA_DIR, copy_test_data, create_node_population, edit_config


class TestMorphHelper:
    def setup_method(self):
        self.nodes = create_node_population(str(TEST_DATA_DIR / "nodes_quaternions.h5"), "default")
        self.morph_path = TEST_DATA_DIR / "morphologies"
        self.test_obj = test_module.MorphHelper(str(self.morph_path), self.nodes)

    def test_biophysical_in_library(self):
        with copy_test_data() as (circuit_copy_path, config_copy_path):
            with edit_config(config_copy_path) as config:
                config["networks"]["nodes"][0]["nodes_file"] = "$NETWORK_DIR/nodes_quaternions.h5"
            nodes_file = circuit_copy_path / "nodes_quaternions.h5"
            with h5py.File(nodes_file, "r+") as h5f:
                data = h5f["nodes/default/0/model_type"][:]
                del h5f["nodes/default/0/model_type"]
                h5f.create_dataset(
                    "nodes/default/0/model_type", data=np.zeros_like(data, dtype=int)
                )
                h5f.create_dataset(
                    "nodes/default/0/@library/model_type",
                    data=np.array(
                        [
                            "biophysical",
                        ],
                        dtype=h5py.string_dtype(),
                    ),
                )

            circuit = Circuit(str(config_copy_path))
            assert isinstance(circuit.nodes["default"].morph, test_module.MorphHelper)

    def test_get_filepath(self):
        node_id = 0
        assert self.nodes.get(node_id, properties="morphology") == "morph-A"

        actual = self.test_obj.get_filepath(node_id)
        expected = self.morph_path / "morph-A.swc"
        assert actual == expected

        actual = self.test_obj.get_filepath(np.int64(node_id))
        assert actual == expected

        actual = self.test_obj.get_filepath(np.uint64(node_id))
        assert actual == expected

        actual = self.test_obj.get_filepath(np.int32(node_id))
        assert actual == expected

        actual = self.test_obj.get_filepath(np.uint32(node_id))
        assert actual == expected

        node_id = CircuitNodeId("default", 0)
        assert self.nodes.get(node_id, properties="morphology") == "morph-A"
        actual = self.test_obj.get_filepath(node_id)
        assert actual == expected

        with pytest.raises(BluepySnapError, match="node_id must be a int or a CircuitNodeId"):
            self.test_obj.get_filepath([CircuitNodeId("default", 0), CircuitNodeId("default", 1)])

        with pytest.raises(BluepySnapError, match="node_id must be a int or a CircuitNodeId"):
            self.test_obj.get_filepath([0, 1])

        with pytest.raises(
            BluepySnapError, match="'neurolucida-asc' is not defined in 'alternate_morphologies'"
        ):
            self.test_obj.get_filepath(0, "asc")

        with pytest.raises(BluepySnapError, match="Unsupported extension: fake"):
            self.test_obj.get_filepath(0, "fake")

    def test_alternate_morphology(self):
        alternate_morphs = {"h5v1": str(self.morph_path)}
        test_obj = test_module.MorphHelper(
            None, self.nodes, alternate_morphologies=alternate_morphs
        )

        node_id = CircuitNodeId("default", 1)
        assert self.nodes.get(node_id, properties="morphology") == "morph-B"
        expected = self.morph_path / "morph-B.h5"
        actual = test_obj.get_filepath(node_id, extension="h5")
        assert actual == expected

        alternate_morphs = {"neurolucida-asc": str(self.morph_path)}
        test_obj = test_module.MorphHelper(
            None, self.nodes, alternate_morphologies=alternate_morphs
        )

        node_id = CircuitNodeId("default", 1)
        assert self.nodes.get(node_id, properties="morphology") == "morph-B"
        expected = self.morph_path / "morph-B.asc"
        actual = test_obj.get_filepath(node_id, extension="asc")
        assert actual == expected

        with pytest.raises(BluepySnapError, match="'morphologies_dir' is not defined in config"):
            node_id = CircuitNodeId("default", 0)
            test_obj.get_filepath(node_id)

    def test_get_morphology(self):
        actual = self.test_obj.get(0)
        assert len(actual.points) == 13
        expected = [
            [0.0, 5.0, 0.0],
            [2.0, 9.0, 0.0],
        ]
        npt.assert_almost_equal(expected, actual.points[:2])
        npt.assert_almost_equal([2.0, 2.0], actual.diameters[:2])

        with pytest.raises(BluepySnapError, match="node_id must be a int or a CircuitNodeId"):
            self.test_obj.get([0, 1])

    def test_get_alternate_morphology(self):
        alternate_morphs = {"h5v1": str(self.morph_path)}
        test_obj = test_module.MorphHelper(
            None, self.nodes, alternate_morphologies=alternate_morphs
        )
        actual = test_obj.get(0, extension="h5")
        assert len(actual.points) == 13
        expected = [
            [0.0, 5.0, 0.0],
            [2.0, 9.0, 0.0],
        ]
        npt.assert_almost_equal(expected, actual.points[:2])
        npt.assert_almost_equal([2.0, 2.0], actual.diameters[:2])

    def test_get_morphology_simple_rotation(self):
        node_id = 0
        # check that the input node positions / orientation values are still the same
        pdt.assert_series_equal(
            self.nodes.positions(node_id),
            pd.Series([101.0, 102.0, 103.0], index=[Node.X, Node.Y, Node.Z], name=0),
        )
        npt.assert_almost_equal(
            self.nodes.orientations(node_id),
            [
                [1, 0.0, 0.0],
                [0.0, 0.0, -1.0],
                [0.0, 1.0, 0.0],
            ],
            decimal=6,
        )

        actual = self.test_obj.get(node_id, transform=True)
        assert len(actual.points) == 13
        # swc file
        # index       type         X            Y            Z       radius       parent
        #   22           2     0.000000     5.000000     0.000000     1.000000           1
        #   23           2     2.000000     9.000000     0.000000     1.000000          22
        # rotation around the x axis 90 degrees counter clockwise (swap Y and Z)
        # x = X + 101, y = Z + 102, z = Y + 103, radius does not change
        expected = [[101.0, 102.0, 108.0], [103.0, 102.0, 112.0]]
        npt.assert_almost_equal(actual.points[:2], expected)
        npt.assert_almost_equal(actual.diameters[:2], [2.0, 2.0])

    def test_get_morphology_standard_rotation(self):
        nodes = create_node_population(str(TEST_DATA_DIR / "nodes.h5"), "default")
        test_obj = test_module.MorphHelper(str(self.morph_path), nodes)

        node_id = 0
        actual = test_obj.get(node_id, transform=True).points

        # check if the input node positions / orientation values are still the same
        pdt.assert_series_equal(
            self.nodes.positions(node_id),
            pd.Series([101.0, 102.0, 103.0], index=[Node.X, Node.Y, Node.Z], name=0),
        )
        npt.assert_almost_equal(
            nodes.orientations(node_id),
            [
                [0.738219, 0.0, 0.674560],
                [0.0, 1.0, 0.0],
                [-0.674560, 0.0, 0.738219],
            ],
            decimal=6,
        )

        assert len(actual) == 13
        expected = [[101.0, 107.0, 103.0], [102.47644, 111.0, 101.65088]]
        npt.assert_almost_equal(actual[:2], expected, decimal=6)



"""Taken from the libsonata lib."""
import h5py
import numpy as np


def write_spikes(filepath):
    population_names = ["default", "default2"]
    timestamps_base = (0.3, 0.1, 0.2, 1.3, 0.7)
    node_ids_base = (1, 2, 0, 0, 2)

    sorting_type = h5py.enum_dtype({"none": 0, "by_id": 1, "by_time": 2})

    with h5py.File(filepath, "w") as h5f:
        h5f.create_group("spikes")
        gpop_spikes = h5f.create_group("/spikes/" + population_names[0])
        gpop_spikes.attrs.create("sorting", data=2, dtype=sorting_type)
        timestamps, node_ids = zip(*sorted(zip(timestamps_base, node_ids_base)))
        gpop_spikes.create_dataset("timestamps", data=timestamps, dtype=np.double)
        gpop_spikes.create_dataset("node_ids", data=node_ids, dtype=np.uint64)

        gpop_spikes2 = h5f.create_group("/spikes/" + population_names[1])
        gpop_spikes2.attrs.create("sorting", data=1, dtype=sorting_type)
        node_ids, timestamps = zip(*sorted(zip(node_ids_base, timestamps_base)))
        gpop_spikes2.create_dataset("timestamps", data=timestamps, dtype=np.double)
        gpop_spikes2.create_dataset("node_ids", data=node_ids, dtype=np.uint64)


def write_soma_report(filepath):
    population_names = ["default", "default2"]
    node_ids = np.arange(0, 3)
    index_pointers = np.arange(0, 4)
    element_ids = np.zeros(3)
    times = (0.0, 1.0, 0.1)
    data = [node_ids + j * 0.1 for j in range(10)]
    string_dtype = h5py.special_dtype(vlen=str)
    with h5py.File(filepath, "w") as h5f:
        h5f.create_group("report")
        gpop_all = h5f.create_group("/report/" + population_names[0])
        ddata = gpop_all.create_dataset("data", data=data, dtype=np.float32)
        ddata.attrs.create("units", data="mV", dtype=string_dtype)
        gmapping = h5f.create_group("/report/" + population_names[0] + "/mapping")

        dnodes = gmapping.create_dataset("node_ids", data=node_ids, dtype=np.uint64)
        dnodes.attrs.create("sorted", data=True, dtype=np.uint8)
        gmapping.create_dataset("index_pointers", data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset("element_ids", data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset("time", data=times, dtype=np.double)
        dtimes.attrs.create("units", data="ms", dtype=string_dtype)

        gpop_soma2 = h5f.create_group("/report/" + population_names[1])
        ddata = gpop_soma2.create_dataset("data", data=data, dtype=np.float32)
        ddata.attrs.create("units", data="mV", dtype=string_dtype)
        gmapping = h5f.create_group("/report/" + population_names[1] + "/mapping")

        dnodes = gmapping.create_dataset("node_ids", data=node_ids, dtype=np.uint64)
        dnodes.attrs.create("sorted", data=True, dtype=np.uint8)
        gmapping.create_dataset("index_pointers", data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset("element_ids", data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset("time", data=times, dtype=np.double)
        dtimes.attrs.create("units", data="ms", dtype=string_dtype)


def write_element_report(filepath):
    population_names = ["default", "default2"]
    node_ids = np.arange(0, 3)
    index_pointers = np.arange(0, 8, 2)
    index_pointers[-1] = index_pointers[-1] + 1
    element_ids = np.array([0, 1] * 3 + [1])

    times = (0.0, 1, 0.1)

    string_dtype = h5py.special_dtype(vlen=str)
    with h5py.File(filepath, "w") as h5f:
        h5f.create_group("report")
        gpop_element = h5f.create_group("/report/" + population_names[0])
        d1 = np.array([np.arange(7) + j * 0.1 for j in range(10)])
        ddata = gpop_element.create_dataset("data", data=d1, dtype=np.float32)
        ddata.attrs.create("units", data="mV", dtype=string_dtype)
        gmapping = h5f.create_group("/report/" + population_names[0] + "/mapping")

        dnodes = gmapping.create_dataset("node_ids", data=node_ids, dtype=np.uint64)
        dnodes.attrs.create("sorted", data=True, dtype=np.uint8)
        gmapping.create_dataset("index_pointers", data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset("element_ids", data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset("time", data=times, dtype=np.double)
        dtimes.attrs.create("units", data="ms", dtype=string_dtype)

        gpop_element2 = h5f.create_group("/report/" + population_names[1])
        d1 = np.array([np.arange(7) + j * 0.1 for j in range(10)])
        ddata = gpop_element2.create_dataset("data", data=d1, dtype=np.float32)
        ddata.attrs.create("units", data="mR", dtype=string_dtype)
        gmapping = h5f.create_group("/report/" + population_names[1] + "/mapping")

        dnodes = gmapping.create_dataset("node_ids", data=node_ids, dtype=np.uint64)
        dnodes.attrs.create("sorted", data=True, dtype=np.uint8)
        gmapping.create_dataset("index_pointers", data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset("element_ids", data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset("time", data=times, dtype=np.double)
        dtimes.attrs.create("units", data="mR", dtype=string_dtype)


if __name__ == "__main__":
    write_spikes("spikes.h5")
    write_soma_report("soma_report.h5")
    write_element_report("compartment_named.h5")



import json
import re
from unittest.mock import patch

import libsonata
import pytest
from numpy.testing import assert_array_equal

import bluepysnap.node_sets as test_module
from bluepysnap.exceptions import BluepySnapError

from utils import TEST_DATA_DIR


class TestNodeSet:
    def setup_method(self):
        self.test_node_sets = test_module.NodeSets.from_file(
            str(TEST_DATA_DIR / "node_sets_file.json")
        )
        self.test_pop = libsonata.NodeStorage(str(TEST_DATA_DIR / "nodes.h5")).open_population(
            "default"
        )

    def test_get_ids(self):
        assert_array_equal(self.test_node_sets["Node2_L6_Y"].get_ids(self.test_pop), [])
        assert_array_equal(self.test_node_sets["double_combined"].get_ids(self.test_pop), [0, 1, 2])

        node_set = self.test_node_sets["failing"]

        with pytest.raises(BluepySnapError, match="No such attribute"):
            node_set.get_ids(self.test_pop)

        assert node_set.get_ids(self.test_pop, raise_missing_property=False) == []


class TestNodeSets:
    def setup_method(self):
        self.test_obj = test_module.NodeSets.from_file(str(TEST_DATA_DIR / "node_sets_file.json"))
        self.test_pop = libsonata.NodeStorage(str(TEST_DATA_DIR / "nodes.h5")).open_population(
            "default"
        )

    def test_init(self):
        assert self.test_obj.content == {
            "double_combined": ["combined", "population_default_L6"],
            "Node2_L6_Y": {"mtype": ["L6_Y"], "node_id": [30, 20, 20]},
            "Layer23": {"layer": [3, 2, 2]},
            "population_default_L6": {"population": "default", "mtype": "L6_Y"},
            "combined": ["Node2_L6_Y", "Layer23"],
            "failing": {"unknown_property": [0]},
        }

    def test_from_string(self):
        res = test_module.NodeSets.from_string(json.dumps(self.test_obj.content))
        assert res.content == self.test_obj.content

    def test_from_dict(self):
        res = test_module.NodeSets.from_dict(self.test_obj.content)
        assert res.content == self.test_obj.content

    def test_update(self):
        # update all keys
        tested = test_module.NodeSets.from_file(str(TEST_DATA_DIR / "node_sets_file.json"))
        res = tested.update(tested)

        # should return all keys as replaced
        assert res == {*self.test_obj}
        assert tested.content == self.test_obj.content

        # actually add a new node set
        additional = {"test": {"test_property": ["test_value"]}}
        res = tested.update(test_module.NodeSets.from_dict(additional))
        expected_content = {**self.test_obj.content, **additional}

        # None of the keys should be replaced
        assert res == set()
        assert tested.content == expected_content

        with pytest.raises(
            BluepySnapError, match=re.escape("Unexpected type: 'str' (expected: 'NodeSets')")
        ):
            tested.update("")

    def test_contains(self):
        assert "Layer23" in self.test_obj
        assert "find_me_you_will_not" not in self.test_obj
        with pytest.raises(BluepySnapError, match="Unexpected type"):
            42 in self.test_obj

    def test_getitem(self):
        assert isinstance(self.test_obj["Layer23"], test_module.NodeSet)

        with pytest.raises(BluepySnapError, match="Undefined node set:"):
            self.test_obj["no-such-node-set"]

    def test_iter(self):
        expected = set(json.loads((TEST_DATA_DIR / "node_sets_file.json").read_text()))
        assert set(self.test_obj) == expected



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Miscellaneous utilities."""

import json
import warnings
from collections.abc import Iterable

import numpy as np

from bluepysnap.circuit_ids import CircuitEdgeId, CircuitNodeId
from bluepysnap.exceptions import (
    BluepySnapDeprecationError,
    BluepySnapDeprecationWarning,
    BluepySnapError,
)
from bluepysnap.sonata_constants import DYNAMICS_PREFIX

# dtypes for the different node and edge ids. We are using np.int64 to avoid the infamous
# https://github.com/numpy/numpy/issues/15084 numpy problem. This type needs to be used for
# all returned node or edge ids.
IDS_DTYPE = np.int64


class Deprecate:
    """Class for the deprecations in BluepySnap."""

    @staticmethod
    def fail(msg=""):
        """Raise a deprecation exception."""
        raise BluepySnapDeprecationError(msg)

    @staticmethod
    def warn(msg=""):
        """Issue a deprecation warning."""
        warnings.warn(msg, BluepySnapDeprecationWarning)


def load_json(filepath):
    """Load JSON from file."""
    with open(filepath, encoding="utf-8") as f:
        return json.load(f)


def is_node_id(node_id):
    """Check if node_id can be a valid node id.

    Returns:
        bool: true is node_id is a int, np.integer or CircuitNodeId. False for anything else.
    """
    return isinstance(node_id, (int, np.integer, CircuitNodeId))


def is_iterable(v):
    """Check if `v` is any iterable (strings are considered scalar and CircuitNode/EdgeId also)."""
    return isinstance(v, Iterable) and not isinstance(v, (str, CircuitNodeId, CircuitEdgeId))


def ensure_list(v):
    """Convert iterable / wrap scalar into list (strings are considered scalar)."""
    if is_iterable(v):
        return list(v)
    else:
        return [v]


def ensure_ids(a):
    """Convert a numpy array dtype into IDS_DTYPE.

    This function is here due to the https://github.com/numpy/numpy/issues/15084 numpy issue.
    It is quite unsafe to the use uint64 for the ids due to this problem where :
    numpy.uint64 + int --> float64
    numpy.uint64 += int --> float64

    This function needs to be used everywhere node_ids or edge_ids are returned.
    """
    return np.asarray(a, IDS_DTYPE)


def add_dynamic_prefix(properties):
    """Add the dynamic prefix to a list of properties."""
    return [DYNAMICS_PREFIX + name for name in list(properties)]


def euler2mat(az, ay, ax):
    """Build 3x3 rotation matrices from az, ay, ax rotation angles (in that order).

    Args:
        az: rotation angles around Z (Nx1 NumPy array; radians)
        ay: rotation angles around Y (Nx1 NumPy array; radians)
        ax: rotation angles around X (Nx1 NumPy array; radians)

    Returns:
        List with Nx3x3 rotation matrices corresponding to each of N angle triplets.

    See Also:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix (R = X1 * Y2 * Z3)
    """
    if len(az) != len(ay) or len(az) != len(ax):
        raise BluepySnapError("All angles must have the same length.")
    c1, s1 = np.cos(ax), np.sin(ax)
    c2, s2 = np.cos(ay), np.sin(ay)
    c3, s3 = np.cos(az), np.sin(az)

    mm = np.array(
        [
            [c2 * c3, -c2 * s3, s2],
            [c1 * s3 + c3 * s1 * s2, c1 * c3 - s1 * s2 * s3, -c2 * s1],
            [s1 * s3 - c1 * c3 * s2, c3 * s1 + c1 * s2 * s3, c1 * c2],
        ]
    )

    return [mm[..., i] for i in range(len(az))]


def quaternion2mat(aqw, aqx, aqy, aqz):
    """Build 3x3 rotation matrices from quaternions.

    Args:
        aqw: w component of quaternions (Nx1 NumPy array; float)
        aqx: x component of quaternions (Nx1 NumPy array; float)
        aqy: y component of quaternions (Nx1 NumPy array; float)
        aqz: z component of quaternions (Nx1 NumPy array; float)

    Returns:
        List with Nx3x3 rotation matrices corresponding to each of N quaternions.

    See Also:
        https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
    """

    def normalize_quaternions(qs):
        """Normalize a bunch of quaternions along axis==1.

        Args:
            qs: quaternions (Nx4 NumPy array; float)

        Returns:
           numpy array of normalized quaternions
        """
        return qs / np.sqrt(np.einsum("...i,...i", qs, qs)).reshape(-1, 1)

    aq = np.dstack([np.asarray(aqw), np.asarray(aqx), np.asarray(aqy), np.asarray(aqz)])[0]
    aq = normalize_quaternions(aq)

    w = aq[:, 0]
    x = aq[:, 1]
    y = aq[:, 2]
    z = aq[:, 3]

    mm = np.array(
        [
            [w * w + x * x - y * y - z * z, 2 * x * y - 2 * w * z, 2 * w * y + 2 * x * z],
            [2 * w * z + 2 * x * y, w * w - x * x + y * y - z * z, 2 * y * z - 2 * w * x],
            [2 * x * z - 2 * w * y, 2 * w * x + 2 * y * z, w * w - x * x - y * y + z * z],
        ]
    )

    return [mm[..., i] for i in range(len(aq))]



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""BBP cell / synapse attribute namespace."""

from bluepysnap.sonata_constants import DYNAMICS_PREFIX, Edge, Node

NODE_TYPES = {"biophysical", "virtual", "astrocyte", "single_compartment"}
EDGE_TYPES = {"chemical", "electrical", "synapse_astrocyte", "endfoot"}


class Cell(Node):
    """Cell property names."""

    ME_COMBO = "me_combo"  #:
    MTYPE = "mtype"  #:
    ETYPE = "etype"  #:
    LAYER = "layer"  #:
    REGION = "region"  #:
    SYNAPSE_CLASS = "synapse_class"  #:
    HOLDING_CURRENT = DYNAMICS_PREFIX + "holding_current"  #:
    THRESHOLD_CURRENT = DYNAMICS_PREFIX + "threshold_current"  #:


class Synapse(Edge):
    """Synapse property names."""

    PRE_GID = Edge.SOURCE_NODE_ID  #:
    POST_GID = Edge.TARGET_NODE_ID  #:

    D_SYN = "depression_time"  #:
    DTC = "decay_time"  #:
    F_SYN = "facilitation_time"  #:
    G_SYNX = "conductance"  #:
    NRRP = "NRRP"  #:
    TYPE = "syn_type_id"  #:
    U_SYN = "u_syn"  #:
    SPINE_LENGTH = "spine_length"  #:

    PRE_SEGMENT_ID = "efferent_segment_id"  #:
    PRE_SEGMENT_OFFSET = "efferent_segment_offset"  #:
    PRE_MORPH_ID = "efferent_morphology_id"  #:

    POST_SEGMENT_ID = "afferent_segment_id"  #:
    POST_SEGMENT_OFFSET = "afferent_segment_offset"  #:
    POST_BRANCH_TYPE = "afferent_section_type"  #:



import bluepysnap._doctools as test_module


class TestClass:
    __test__ = False

    def __init__(self):
        """TestClass"""
        self.foo_val = 0
        self.bar_val = 1

    def foo(self):
        """foo function for TestClass"""
        return self.foo_val

    def bar(self):
        """bar function for TestClass

        Returns:
            TestClass: return a TestClass
        """
        return self.bar_val

    def foo_bar(self):
        pass


class TestClassA(
    TestClass,
    metaclass=test_module.DocSubstitutionMeta,
    source_word="TestClass",
    target_word="TestClassA",
):
    """New class with changed docstrings."""


class TestClassB(
    TestClass,
    metaclass=test_module.DocSubstitutionMeta,
    source_word="TestClass",
    target_word="TestClassB",
):
    """New class with changed docstrings."""


class TestClassC(TestClassA):
    def bar(self):
        """Is overrode correctly"""
        return 42


def test_DocSubstitutionMeta():
    default = TestClass()
    tested = TestClassA()
    assert tested.foo.__doc__ == default.foo.__doc__.replace("TestClass", "TestClassA")
    expected = default.bar.__doc__.replace("TestClass", "TestClassA")
    assert tested.bar.__doc__ == expected
    assert tested.foo_bar.__doc__ is None
    assert tested.__dict__ == default.__dict__

    # do not override the mother class docstrings
    tested = TestClassB()
    assert tested.foo.__doc__ == default.foo.__doc__.replace("TestClass", "TestClassB")
    expected = default.bar.__doc__.replace("TestClass", "TestClassB")
    assert tested.bar.__doc__ == expected
    assert tested.foo_bar.__doc__ is None
    assert tested.__dict__ == default.__dict__

    # I can inherit from a class above
    tested = TestClassC()
    assert tested.foo.__doc__ == TestClassA.foo.__doc__
    assert tested.bar.__doc__ == "Is overrode correctly"
    assert tested.bar() == 42
